from pathlib import Path

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction, QFileDialog, QVBoxLayout, QWidget

import pixie16

from ..widgets import BaseApp
from .main_widget import CoincidencePlot


class CoincidenceApp(BaseApp):
    """An app that display arbitrary fast trigger positions and all coincidence signals that result.

    It also includes the internal delay from the calculation of the coincidences (added to the validation window).
    """

    def __init__(self, app=None):
        super().__init__(name="Pixie16 Coincidence Visualizer", app=app)

        self.settings_filename = None
        self.settings = None
        self.channels = [0, 1, 2]

        FileButton = QAction(QIcon.fromTheme("document-open"), "Open", self)
        FileButton.setShortcut("Ctrl+O")
        FileButton.setStatusTip("Open settings file")
        FileButton.triggered.connect(lambda x: self.open_file())
        self.file_menu.insertAction(self.file_menu.actions()[0], FileButton)

        self.main_widget = QWidget(self)
        self.vbox = QVBoxLayout()

        self.data = CoincidencePlot(main=self)

        self.info_bar.setText(
            "Change values on the right or click on the graph to change the position of the fast trigger location."
        )

        self.vbox.addLayout(self.data)
        self.vbox.addWidget(self.info_bar)

        self.main_widget.setLayout(self.vbox)
        self.main_widget.setFocus()

        self.setCentralWidget(self.main_widget)

        self.show()

    def open_file(self, filename=None) -> None:
        if filename is None or not isinstance(filename, (str, Path)):
            self.settings_filename, _ = QFileDialog.getOpenFileName(
                self, "Open file", str(self.data_path), "Settings files (*.set)"
            )
        else:
            self.settings_filename = filename

        # check if cancel was hit in QFileDialog
        if self.settings_filename == "":
            self.settings_filename = None
            return

        self.settings_filename = Path(self.settings_filename)
        self.settings = pixie16.read.load_settings(self.settings_filename, 0)

        channels = [i for i, x in enumerate(self.settings.MultiplicityMaskL) if x != 0]
        if len(channels) < 1:
            print("[Warning] no coincidence setup detected, not loading settings")
            return

        # get all channels that are involved
        all_channels = []
        for c in channels:
            tmp = self.settings.MultiplicityMaskL[c]
            tmp = f"{tmp:016b}"
            for i, n in enumerate(reversed(tmp)):
                if int(n):
                    all_channels.append(i)
        self.channels = list(set(all_channels))
        self.data.multiplicity = [
            self.settings.MultiplicityMaskL[c] for c in self.channels
        ]

        self.data.channels = self.channels

        self.data.update_channels()

        for i, c in enumerate(self.channels):
            values = {
                "Trigger Delay": self.settings.FtrigoutDelay[c],
                "Trigger Stretch": self.settings.FastTrigBackLen[c],
                "Validation Delay": self.settings.ExternDelayLen[c],
                "Validation Stretch": self.settings.ChanTrigStretch[c],
            }
            self.data.channel_widgets[i].vstack.value = values



from pathlib import Path

import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

import pixie16


@pytest.fixture(scope="module")
def full_list_mode_binary_file():
    return Path(__file__).parent / "pixie16_binary_data-full.bin"


@pytest.fixture(scope="module")
def split_list_mode_binary_files():
    files = [
        "pixie16_binary_data-00.bin",
        "pixie16_binary_data-01.bin",
        "pixie16_binary_data-02.bin",
        "pixie16_binary_data-03.bin",
        "pixie16_binary_data-04.bin",
    ]

    path = Path(__file__).parent
    return [path / f for f in files]


@pytest.fixture(scope="module")
def list_mode_data_events():
    path = Path(__file__).parent
    events_file = path / "pixie16-events.parquet"
    data = pd.read_parquet(events_file)
    # fix some data types
    for column in ["pileup", "trace_flag"]:
        data[column] = data[column].astype(bool)
    for column in ["Esum_trailing", "Esum_leading", "Esum_gap", "baseline"]:
        data[column] = data[column].replace([0], -1)
    return data


class TestReadListModeData:
    def test_read_full_binary_file(
        self, full_list_mode_binary_file, list_mode_data_events
    ):
        read_events = pixie16.read.read_list_mode_data([full_list_mode_binary_file])

        to_drop = list(set(read_events.columns) - set(list_mode_data_events.columns))
        read_events.drop(to_drop, axis=1, inplace=True)

        assert_frame_equal(read_events, list_mode_data_events)

    def test_read_split_binary_file(
        self, split_list_mode_binary_files, list_mode_data_events
    ):
        read_events = pixie16.read.read_list_mode_data(split_list_mode_binary_files)

        to_drop = list(set(read_events.columns) - set(list_mode_data_events.columns))
        read_events.drop(to_drop, axis=1, inplace=True)

        assert_frame_equal(read_events, list_mode_data_events)



"""Python interface to controlling the pixie16.

Based on XIA's C-library from their SDK, we try to provide a more
pythonic interface, make use of a config file to store different
configurations, and supply addtional functionality.

The config file (which is a requirement) stores some of the parameter
for booting the pixie as well as the main output directory. Therefore,
our boot and init commands take less parameters then their SDK version.

Once the `init_system` function was called, we also cache the list of
modules. We then provide functions that don't need the addtional input
of curent modules.

We provide some functions to modify channel and module parameters
directly on the pixie, as well as, setting the run time and sync mode
or enabling taking traces. Furthermore, there we provide function to
start/stop acquisition and get information on all modules and save
this to a json file.

We also provide our own mechanism of changing settings by providing a
dictionary with name of the settings as key in the dictioinary and the
settings value only provided for the channels in use. Since the
pixie16 provides many settings by manipulating bits in a 32bit
integer, we provide functions that help with this and one can access
bits by name (inspired by the names in the manual).  We also provide
the option to overwrite some of the calculated values such as
`PAFlength` or have them calculated automatically.

At the moment pypixie16 only supports Windows, but other platforms
should be easy to add.

"""
from contextlib import contextmanager
from collections import OrderedDict
import datetime
import json
import logging
from pathlib import Path
import sys
import tempfile
import time
from typing import Iterable, List, Tuple, Dict, Union, Optional, ByteString

from importlib.resources import files
import numpy as np
from rich.progress import Progress
from rich import print

from . import config
from . import C_library_sdk as SDK
from . import variables
from .read import Settings, Stats, CrateSettings

# set up logging
log = logging.getLogger(__name__)

module_list: List[int] = [2]  # default setting, overwritten by init_and_boot

SETTINGS = variables.settings
SETTINGS_NAME_CHANNEL = []
SETTINGS_NAME_MODULE = []

BIT_PARAMETERS = {
    "FastTrigSelect": "CHANNEL_CSRA",
    "ModValSignal": "CHANNEL_CSRA",
    "GoodChannel": "CHANNEL_CSRA",
    "ChanValSignal": "CHANNEL_CSRA",
    "RejectIfFull": "CHANNEL_CSRA",
    "Polarity": "CHANNEL_CSRA",
    "EnableVeto": "CHANNEL_CSRA",
    "CaptureHistogram": "CHANNEL_CSRA",
    "CaptureTrace": "CHANNEL_CSRA",
    "EnableQDC": "CHANNEL_CSRA",
    "EnableCFD": "CHANNEL_CSRA",
    "EnableModVal": "CHANNEL_CSRA",
    "CaptureSums": "CHANNEL_CSRA",
    "EnableChannelVal": "CHANNEL_CSRA",
    "Gain": "CHANNEL_CSRA",
    "RejectPileup": "CHANNEL_CSRA",
    "SkipLargePulses": "CHANNEL_CSRA",
    "GroupTrigSignal": "CHANNEL_CSRA",
    "ChannelVetoSignal": "CHANNEL_CSRA",
    "ModVetoSignal": "CHANNEL_CSRA",
    "ExtTimestamps": "CHANNEL_CSRA",
    "BackplanePullup": "MODULE_CSRB",
    "Director": "MODULE_CSRB",
    "ChassisMaster": "MODULE_CSRB",
    "GlobalFastTrigger": "MODULE_CSRB",
    "ExternalTrigger": "MODULE_CSRB",
    "ExternalInhibit": "MODULE_CSRB",
    "DistributeClocks": "MODULE_CSRB",
    "SortEvents": "MODULE_CSRB",
    "ConnectFastTriggerBP": "MODULE_CSRB",
}
CHANNEL_PARAMETERS = SDK.valid_channel_parameter_names + [
    x for x in BIT_PARAMETERS if BIT_PARAMETERS[x] == "CHANNEL_CSRA"
]
MODULE_PARAMETERS = SDK.valid_module_parameter_names + [
    x for x in BIT_PARAMETERS if BIT_PARAMETERS[x] == "MODULE_CSRB"
]


for setting_name, (startpos, length) in SETTINGS.items():
    if length == 16:
        SETTINGS_NAME_CHANNEL.append(setting_name)
    else:
        SETTINGS_NAME_MODULE.append(setting_name)


def init(modules: List[int], offline_mode: bool = False, verbose: bool = False) -> None:
    """Initialize the system.

    Calls init_system in the XIA library and saves the list of modules, so that we can use it as a default in later XIA library calls.

    Parameters
    ----------
    modules
        list of modules (slot numbers, e.g. [2,3])
    offline_mode
        Run with the pixie present or not
    verbose
        print some extra information
    """
    global module_list
    module_list = modules
    if not SDK.load_library():
        print(
            "[orange3]WARNING[/] Mocking the pixie library (useful for testing when the library is not installed)"
        )

    SDK.init_system(module_list, offline_mode, verbose)


def boot(
    section_name: Optional[str] = "default",
    modules: Optional[List[int]] = None,
    boot_pattern: int = 0x7F,
    verbose: bool = False,
) -> None:
    """Boot modules with boot_pattern.

    Parameters
    ----------
    section_name
        name for Firmware section in config file that defines the FPGA code, etc that should be used.
    modules
        Boot certain modules (module list given in positions, i.e. not slots), if `None` use modules defined in `init` call.
    boot_pattern
        Bit pattern defining what gets booted.
    verbose
        Print some extra information if True
    """
    FPGAcom, FPGAsp, dsp_code, dsp_var, setting = config.load_config(section_name)

    if modules:
        for m in modules:
            SDK.boot_module(
                setting, FPGAcom, FPGAsp, dsp_code, dsp_var, m, boot_pattern, verbose
            )
    else:
        SDK.boot_module(
            setting,
            FPGAcom,
            FPGAsp,
            dsp_code,
            dsp_var,
            len(module_list),
            boot_pattern,
            verbose,
        )


def init_and_boot(
    offline_mode: bool = False,
    section_name: Optional[str] = "default",
    modules: Optional[List[int]] = None,
    boot_pattern: int = 0x7F,
    verbose: bool = False,
) -> None:
    """Init and boot in one function call.

    Parameters
    ----------
    modules
        list of modules (slot numbers, e.g. [2,3])
    offline_mode
        Run with the pixie present or not
    section_name
        name for Firmware section in config file that defines the FPGA code, etc that should be used.
    boot_pattern
        Bit pattern defining what gets booted.
    verbose
        Print some extra information if True
    """
    if modules is None:
        modules = [2]
    init(modules, offline_mode, verbose)
    boot(section_name, boot_pattern=boot_pattern, verbose=verbose)


def adjust_offsets(modules: Optional[Iterable[int]] = None) -> None:
    """Adjust offsets for all channels in the list of module".

    Parameters
    ----------
    modules
        Iterable of slots with modules (e.g. [2, 3]), uses modules defined in init call otherwise
    """
    if modules is None:
        modules = module_list

    for m, _ in enumerate(modules):
        SDK.adjust_offsets(m)


def set_sync_mode(modules: Optional[Iterable[int]] = None) -> None:
    """Turn on sync mode for every module.

    Parameters
    ----------
    modules
        Iterable of slots with modules (e.g. [2, 3]), uses modules defined in init call otherwise
    """
    if modules is None:
        modules = module_list

    for m, _ in enumerate(modules):
        SDK.write_single_module_parameter("SYNCH_WAIT", 1, m)
        SDK.write_single_module_parameter("IN_SYNCH", 0, m)


def set_channel_parameter(name: str, value, module: int, channel: int):
    """Set and read back a parameter for a channel on the pixie16.

    Parameters
    ----------
    name
        The name of the parameter
    value
        The new value
    module
        The module number
    channel
        The channel number

    Returns
    -------
        The new setting
    """
    SDK.write_single_channel_parameter(name, value, module, channel)
    return SDK.read_single_channel_parameter(
        name, module, channel, convert_bit_pattern=False
    )


def set_module_parameter(name: str, value, module: int):
    """Set and read back a parameter for a module on the pixie16.

    Parameters
    ----------
    name
        The setting name
    value
        The new settings value
    module
        The module number (starting with 0)

    Returns
    -------
        The new setting
    """
    SDK.write_single_module_parameter(name, value, module)
    return SDK.read_single_module_parameter(name, module)


def set_run_time(runtime, module):
    """Sets the run time for the next run on the pixie16.

    Parameters
    ----------
    runtime
        time in seconds
    module
        module number starting at 0
    """
    return set_module_parameter("HOST_RT_PRESET", runtime, module)


def set_traces(module: int, channel: int, status: bool) -> None:
    """Turn on/off taking traces for a certain channel in a specific module on the pixie16.

    Parameters
    ----------
    module
        The module number (starting at 0)
    channel
        The channel number (0-15)
    status
        True: record traces, False: do not record traces
    """
    channel_setting = SDK.read_single_channel_parameter(
        "CHANNEL_CSRA", module, channel, convert_bit_pattern=True
    )
    # we need to set bit eight
    # in python we can address the last element as -1, which is bit 0, so bit 8 is -9
    channel_setting[-9] = status
    SDK.write_single_channel_parameter("CHANNEL_CSRA", channel_setting, module, channel)


def empty_fifo(module: int) -> None:
    """Read all data in a fifo of a specific module and discard it.

    Parameters
    ----------
    module
        The module number (starting at 0)
    """
    while number_of_words := SDK.check_external_FIFO_status(module):
        SDK.read_data_from_external_FIFO(module, number_of_words)


def empty_all_fifos() -> None:
    """Read data from all fifos and discard it.

    Uses the module list defined in `init`.
    """
    for m, _ in enumerate(module_list):
        empty_fifo(m)


def start_histogram_run() -> None:
    """Start a histogram run.

    Uses the module list defined during `init`.
    """
    for m, _ in enumerate(module_list):
        SDK.start_histogram_run(m)


def start_list_mode_run() -> None:
    """Start a list mode run.

    Uses the module list defined during `init`.
    """
    for m, _ in enumerate(module_list):
        SDK.start_list_mode_run(m)


def check_run_status() -> bool:
    """Checks the run status of all known modules."""

    status = True
    for m, _ in enumerate(module_list):
        status = status & SDK.check_run_status(m)
    return status


def get_stats() -> List[Stats]:
    """Get statistics for all modules."""
    return [
        Stats.from_dict(SDK.read_statistics_from_module(m))
        for m, _ in enumerate(module_list)
    ]


def save_stats_to_json(filename: Path) -> None:
    """Get statistics for all modules and save as json to file.

    Parameters
    ----------
    filename
        name to save the settings to
    """
    stats = get_stats()
    # convert to something that can be saved as json (our Stats class doesn't work)
    out = [dict(x) for x in stats]
    if filename.is_file():
        log.error(f"Filename {filename} exists. Not overwriting!!")
        return

    with filename.open("w") as f:
        json.dump(out, f, indent=2)


def end_run() -> None:
    """Calls end_run on all the modules."""
    for m, _ in enumerate(module_list):
        SDK.end_run(m)


def exit_system() -> None:
    """Calls exit_system on all the modules."""
    for m, _ in enumerate(module_list):
        SDK.exit_system(m)


def read_list_mode_fifo(
    check: bool = True, threshold: int = 64, modules: Optional[Iterable[int]] = None
) -> List[np.ndarray]:
    """Reads data from pixies FIFO across all modules defined in pixie16.control.modules

    Parameters
    ----------
    check
        If True, check first if there is enough data that should be read.
        Otherwise always read all data.
    threshold
        Minimum number of data to read when `check` is active.
    modules
        List of modules (slot numbers, e.g. [2]). Uses modules defined in `init` if None.

    Returns
    -------
    List[np.ndarray]
        List with data as a numpy array of 32 bit unsigned integers for each module.
    """
    if modules is None:
        modules = module_list

    do_read = True
    if check:
        do_read = False
        for i, _ in enumerate(modules):
            number_of_words = SDK.check_external_FIFO_status(i)
            if number_of_words > threshold:
                do_read = True
                break

    output = []
    if do_read:
        for i, _ in enumerate(modules):
            number_of_words = SDK.check_external_FIFO_status(i)
            if number_of_words > 0:
                data = SDK.read_data_from_external_FIFO(i, number_of_words)
            else:
                data = np.array([], dtype=np.uint32)
            output.append(data)

    return output


def run_list_mode(filename: Optional[str] = None, runtime: int = 5) -> None:
    """Run the pixie16 in list mode.

    Start and stop a list mode run. The module needs to be
    initialized.  Data will be written to a file. If the filename
    doesn't end with '.bin' the ending will be added. We use the same
    dataformat as the pixie uses internally.  We also add a '000' or
    higher number before the '.bin' file ending automatically to avoid
    overiding an existing file.  The file gets placed in a the
    directory specified in the config file and within that directory
    in a subdirectory of the form YYYY-MM-DD, which gets created if it
    doesn't exist.

    Parameters
    ----------
    filename
       the filename
    runtime
       The time to take data for in seconds
    """
    YYYYMMDD = datetime.datetime.today().strftime("%Y-%m-%d")
    if filename is None:
        filename = "pixie16-data"

    # remove .bin, will add it back in a bit
    if filename.endswith(".bin"):
        filename = filename[:-4]
    # check if filename has 3 digits at the end
    number = filename[-3:]
    try:
        number = int(number) + 1
        filename = filename[:-3]
    except ValueError:
        number = 0
    if number > 999:
        print("list-mode-data: filenumber too large. Use a new filename....existing!")
        sys.exit()

    filename = f"{filename}{number:03d}.bin"

    if not filename.startswith(YYYYMMDD):
        filename = f"{YYYYMMDD}-{filename}"
    # add correct directory
    datadir = config.config_get_parameters("Data", "datadir", verbose=False)
    if datadir is None:
        datadir = Path(".")
        print(
            f"[orange3]WARNING[/] Could not find datadir in Data section in ini-file {config.inifile}."
        )
        print(f"[orange3]WARNING[/] Using {datadir.absolute()} to save data.")
    filename = datadir / YYYYMMDD / filename
    # make sure directory exists
    filename.parent.mkdir(parents=True, exist_ok=True)

    if filename.exists():
        print(f"filename {filename} already exists...exiting")
        return

    with filename.open("wb") as outfile:
        start_list_mode_run()
        start = time.time()
        stop = start + runtime

        while time.time() < stop:
            data = read_list_mode_fifo()
            for d in data:
                d.newbyteorder("S").tofile(outfile)

        end_run()
        time.sleep(0.4)

        # read final data
        data = read_list_mode_fifo(check=False)
        for d in data:
            d.newbyteorder("S").tofile(outfile)


def reset_coincidence_setting(channels: Iterable[Tuple[int, int]]) -> None:
    """Reset all setting in regards to coincedence mode.

    Also, unsets capturing traces, etc.

    This is usefule before an MCA run for example to just get raw channel spectra.

    Parameters
    ----------
    channels
        List of (module, channel) tuples
    """
    settings = OrderedDict(
        {
            "channels": channels,
            "MultiplicityMaskL": 0,
            "MultiplicityMaskH": 0,
            "TrigConfig0": 0,
            "TrigConfig1": 0,
            "TrigConfig2": 0,
            "TrigConfig3": 0,
            "CaptureTrace": False,
            "CaptureHistogram": True,
            "CaptureSums": False,
            "FastTrigSelect": "group",
            "EnableModVal": False,
            "EnableChannelVal": False,
            "GroupTrigSignal": "local",
            "RejectPileup": "pileup",
            "RejectIfFull": False,
        }
    )
    change_setting_from_dict(settings)


def enable_trace_settings(
    channels: Iterable[Tuple[int, int]], disable_CFD: bool = True
) -> None:
    """Enable traces, histograms, and sums.

    Also turn CFD settings off (or optionally leave unchanged).  These
    data are useful when trying to analyze traces and optimize
    settings offline by trying out different settings values in
    software, for example, using pypixie16's binary-browser.

    Parameters
    ----------
    channels
        List of (modules, channel) tuples
    disable_CFD
        Option to disable CFD for the listed channels

    """
    settings = OrderedDict(
        {
            "channels": channels,
            "CaptureTrace": True,
            "CaptureHistogram": True,
            "CaptureSums": True,
        }
    )
    change_setting_from_dict(settings)

    if disable_CFD:
        print("[yellow]INFO[/] traces: turned CFD off")
        settings = OrderedDict({"channels": channels, "EnableCFD": False})
        change_setting_from_dict(settings)


def read_histograms(module_channels: Iterable[Tuple[int, int]]) -> List[np.ndarray]:
    """Read histograms from several channels.

    Uses (module, channel) tuples, e.g. [(0, 8), (0, 3)] to read histograms.

    Parameters
    ----------
    module_channels
        List of tuples of (module, channel) to define which histograms we want to read

    Returns
    -------
    List[np.ndarray]
        List of numpy arrays with the histograms
    """
    return [SDK.read_histogram_from_module(m, c) for m, c in module_channels]


def take_MCA_spectra(
    channels: Iterable[Tuple[int, int]],
    duration: float,
    verbose: bool = True,
    position: int = 0,
) -> List[np.ndarray]:
    """Takes MCA spectra for a certain time on the specified channels.

    This does the data acquisition and returns the data.

    Parameters
    ----------
    channels
         list of (modules, channel number) tuples
    duration
         MCA time in seconds

    Returns
    -------
    list(nd.ndarray)
         List of numpy arrays. One for each channel.
    """
    settings = OrderedDict(
        {
            "channels": channels,
            "HOST_RT_PRESET": duration,
        }
    )
    change_setting_from_dict(settings)

    modules = list({x for x, y in channels})  # unique list of modules

    # take MCA spectra
    if verbose:
        print("[yellow]INFO[/] Taking MCA spectrum", end="", flush=True)

    start_histogram_run()
    sys.stdout.flush()
    time.sleep(0.5)

    start = time.time()
    keep_running = True
    with Progress() as progress:
        task = progress.add_task("[red] MCA", total=duration)

        while keep_running and (time.time() - start < duration):
            keep_running = False
            for m, _ in enumerate(modules):
                r = SDK.check_run_status(m)
                sys.stdout.flush()
                if r == 1:
                    keep_running = True
            time.sleep(1)
            dt = time.time() - start
            progress.update(task, advance=dt)
        progress.update(task, completed=duration)
    print()
    end_run()

    return read_histograms(channels)


def take_list_mode_data(
    duration: float,
    modules: Optional[Iterable[int]] = None,
) -> Dict[int, ByteString]:
    """Take list mode data for a certain time.

    It also stops at 1 Gb of raw data to avoid too much memory use.
    If you want to take mor data, you need to use another mechanism
    and write the data to disk more often.

    Parameters
    ----------
    duration
       Length in seconds of how long data is acquired for
    modules
       List of modules, use the ones defined during `init` if None.

    Returns
    -------
    dict[int, ByteString]
       Dictionary with the raw binary data in it. Each module gets is a key
       in the dictionary and the values are the bytestring.
    """
    if modules is None:
        modules = module_list

    start_list_mode_run()

    # initialize output module
    raw_data = {i: b"" for i, _ in enumerate(modules)}

    start = time.time()
    raw_data_length = 0

    with Progress() as progress:
        task = progress.add_task("[red] Traces", total=duration)

        while (time.time() - start < duration) and (raw_data_length < 1e9):
            data = read_list_mode_fifo(threshold=32 * 1024)
            for i, d in enumerate(data):
                if d is not None:
                    raw_data[i] += d.tobytes()
                    raw_data_length += len(d)
            dt = time.time() - start
            progress.update(task, advance=dt)
        progress.update(task, completed=duration)

    end_run()
    time.sleep(0.5)

    # read remainig data from pixie
    tmp = read_list_mode_fifo(check=False)
    if not tmp:
        return raw_data

    for i, d in enumerate(tmp):
        if d is not None:
            raw_data[i] += d.tobytes()

    return raw_data


def save_settings_to_json(filename: Path):
    """Save the current settings to a file."""
    SDK.save_dsp_parameters_as_json(filename)


def read_settings_from_json(filename: Path):
    """Reads the settings from a new-style json settings file."""
    with filename.open("r") as f:
        data = json.load(f)
    return data


def read_settings() -> CrateSettings:
    """Read the current setting into memory.

    Reads the setting into a temporary json file using XIA's library and then parse that file.

    Returns
    -------
       A dictonary of setting data
    """
    # note: using tempfile in a context manager has problems on windows when writing and
    # then reading from the same file
    f = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
    filename = Path(f.name)
    SDK.save_dsp_parameters_as_json(filename)
    f.seek(0)
    data = json.load(f)
    f.close()
    filename.unlink()
    if data is None:
        print("[red]Error[/] Cannot read settings from pixie")
        return CrateSettings([])
    return CrateSettings([Settings.from_dict(x) for x in data])


def read_dummy_settings(modules) -> CrateSettings:
    """Read the current setting into memory.

    Reads settings from a given file to be used as dummy data.
    Needed, for example, for tests.

    Returns
    -------
       A dictonary of setting data
    """
    with files("pixie16.data").joinpath("default.json").open("r") as f:
        data = json.load(f)
    return CrateSettings([Settings.from_dict(x) for x in data])


def write_settings_from_file(filename: Path) -> None:
    """Loads a setting file into the pixie16.

    Parameters
    ----------
    filename
        settings file to load

    """
    SDK.load_dsp_parameters_from_file(filename)


def write_settings(data: CrateSettings) -> None:
    """Writes a setting object to the pixie16.

    Parameters
    ----------
    data
        Object created by `read_settings`.
    """
    # note: using tempfile in a context manager has problems on windows when writing and
    # then reading from the same file
    f = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
    filename = Path(f.name)
    with filename.open("w") as f:
        json.dump([dict(x) for x in data], f, indent=2)
    SDK.load_dsp_parameters_from_file(filename)
    filename.unlink()


@contextmanager
def temporary_settings() -> None:
    """Making it easy to temporary change settings in the pixie16.

    A context manager that will remember the current setting, execute
    some code that can change those settings, and then reset the
    setting back to the original.

    It saves the settings of all modules defined in the `init` call.
    """
    current_settings = read_settings()

    try:
        yield None  # this is were the code gets executed
    finally:
        write_settings(current_settings)


def change_setting_from_dict(settings: dict, call_pixie: Optional[bool] = True):
    """Update settings in the pixie16.

    Takes a dictionary with setting names as keys and setting values

    The dictionary must also contain an entry called 'channels' that
    list all channels that should be set. Channels should be pairs in
    the form (module, channel).

    The values in the dictionary can either be a single value, in which case the value
    will be used for all channels or a list of the same length as `channels` in which case
    each (module, channel) combination will get its own value.
    """
    assert (
        "channels" in settings
    ), "The settings dictionary needs an entry listing called 'channels' the channels"
    channels = settings.pop("channels")

    for c in channels:
        assert isinstance(
            c, (list, tuple)
        ), "Setting dictionary: each channel must be a list or tuple"
        assert (
            len(c) == 2
        ), "Setting dictionary: each channel must have two entries (modules, channel)"

    modules = {x[0] for x in channels}

    out = []  # this list is mainly for testing
    for name, value in settings.items():
        if name in CHANNEL_PARAMETERS:
            # convert single values to lists that matches the number of channels
            if isinstance(value, (int, float, str)):
                value_lst = [value] * len(channels)
            else:
                value_lst = value
            # set each value in the list
            for (module, channel), ch_value in zip(channels, value_lst):
                # if the parameter is one that we defined for a bit value, we need to
                # read the current value, change the single bit before we write it
                if name in BIT_PARAMETERS:
                    pixie_param_name = BIT_PARAMETERS[name]
                    if call_pixie:
                        bit_param_val = SDK.read_single_channel_parameter(
                            pixie_param_name, module, channel, convert_bit_pattern=True
                        )
                    else:
                        bit_param_val = [False] * 32
                    ch_value_out = _change_bit_parameter(name, ch_value, bit_param_val)
                    name_out = pixie_param_name
                else:
                    ch_value_out = ch_value
                    name_out = name
                if call_pixie:
                    SDK.write_single_channel_parameter(
                        name_out, ch_value_out, module, channel
                    )
                out.append(["channel", name_out, ch_value, module, channel])
        elif name in MODULE_PARAMETERS:
            # for a module parameter things are a bit simpler, since we don't
            # have to scale things from a single value
            for module in modules:
                # we still need to handle some parameters that have bit values
                # for which we defined our own names
                if name in BIT_PARAMETERS:
                    pixie_param_name = BIT_PARAMETERS[name]
                    if call_pixie:
                        bit_param_val = SDK.read_single_module_parameter(
                            pixie_param_name, module, convert_bit_pattern=True
                        )
                    else:
                        bit_param_val = [False] * 32
                    value = _change_bit_parameter(name, value, bit_param_val)
                    name = pixie_param_name
                if call_pixie:
                    SDK.write_single_module_parameter(name, value, module)
                out.append(["module", name, value, module])

        else:
            print(f"[red]Error[/] {name} not a valid Pixie16 parameter name.")
    return out


def _change_bit_parameter(
    name: str, value: Union[bool, int, str], bit_list: List[bool]
) -> List[bool]:
    """Internal function to adjust single bits in MODULE_CSRB and CHANNEL_CSRA.

    This sets the correct bits for certain settings, so that the user doesn't have to
    figure out which bits to set in an integer. To make this easier, we invented new
    settings names listed below that correspond to certain bits. The table below will
    list the valid parameters.

    .. list-table:: Functions
       :widths: 25 25
       :header-rows: 1

       * - Setting Name
         - Parameter
         - bit
         - possible values
       * - BackplanePullup
         - MODULE_CSRB
         - 0
         - True/False
       * - Director
         - MODULE_CSRB
         - 4
         - True/False
       * - ChassisMaster
         - MODULE_CSRB
         - 6
         - True/False
       * - GlobalFastTrigger
         - MODULE_CSRB
         - 7
         - True/False
       * - ExternalTrigger
         - MODULE_CSRB
         - 8
         - True/False
       * - ExternalInhibit
         - MODULE_CSRB
         - 10
         - True/False
       * - DistributeClocks
         - MODULE_CSRB
         - 11
         - True/False
       * - SortEvents
         - MODULE_CSRB
         - 12
         - True/False
       * - ConnectFastTriggerBP
         - MODULE_CSRB
         - 13
         - True/False
       * - FastTrigSelect
         - CHANNEL_CSRA
         - 0
         - external(=True)/group(=False)
       * - ModValSignal
         - CHANNEL_CSRA
         - 1
         - modgate(=True)/global(=False)
       * - GoodChannel
         - CHANNEL_CSRA
         - 2
         - True/False
       * - ChanValSignal
         - CHANNEL_CSRA
         - 3
         - channelgate(=True)/channelvalidation(=False)
       * - RejectIfFull
         - CHANNEL_CSRA
         - 4
         - True/False
       * - Polarity
         - CHANNEL_CSRA
         - 5
         - positive(=True)/negative(=False)
       * - EnableVeto
         - CHANNEL_CSRA
         - 6
         - True/False
       * - CaptureHistogram
         - CHANNEL_CSRA
         - 7
         - True/False
       * - CaptureTrace
         - CHANNEL_CSRA
         - 8
         - True/False
       * - EnableQDC
         - CHANNEL_CSRA
         - 9
         - True/False
       * - EnableCFD
         - CHANNEL_CSRA
         - 10
         - True/False
       * - EnableModVal
         - CHANNEL_CSRA
         - 11
         - True/False
       * - CaptureSums
         - CHANNEL_CSRA
         - 12
         - True/False
       * - EnableChannelVal
         - CHANNEL_CSRA
         - 13
         - True/False
       * - Gain
         - CHANNEL_CSRA
         - 14
         - 2.5(=True)/0.625(=False)
       * - RejectPileup
         - CHANNEL_CSRA
         - 15, 16
         - single(b15=1, b16=0)/pileup-only(b15=1, b16=1)/pileup(b15=0, b16=1)/all(b15=0, b16=0)
       * - SkipLargePulses
         - CHANNEL_CSRA
         - 17
         - True/False
       * - GroupTrigSignal
         - CHANNEL_CSRA
         - 18
         - external(=True)/local(=False)
       * - ChannelVetoSignal
         - CHANNEL_CSRA
         - 19
         - channel(=True)/front(=False)
       * - ModVetoSignal
         - CHANNEL_CSRA
         - 20
         - channel(=True)/front(=False)
       * - ExtTimestamps
         - CHANNEL_CSRA
         - 21
         - True/False
    """
    # Module parameters (currently onlChanValSignaly MODULE_CSRB)
    if name == "BackplanePullup":
        # value: True = connect backplane to pullup resistor
        assert isinstance(
            value, bool
        ), "BackplanePullup needs to be either True or False"
        bit_list[31 - 0] = value
    elif name == "Director":
        # value: True = set to director
        assert isinstance(value, bool), "Director needs to be either True or False"
        bit_list[31 - 4] = value
    elif name == "ChassisMaster":
        # value: True = set to chassis master
        assert isinstance(value, bool), "ChassisMaster needs to be either True or False"
        bit_list[31 - 6] = value
    elif name == "GlobalFastTrigger":
        # value: True = select global fast trigger source
        assert isinstance(
            value, bool
        ), "GlobalFastTrigger needs to be either True or False"
        bit_list[31 - 7] = value
    elif name == "ExternalTrigger":
        # value: True = select external trigger source
        assert isinstance(
            value, bool
        ), "ExternalTrigger needs to be either True or False"
        bit_list[31 - 8] = value
    elif name == "ExternalInhibit":
        # value: True = use inhibit
        assert isinstance(
            value, bool
        ), "ExternalInhibit needs to be either True or False"
        bit_list[31 - 10] = value
    elif name == "DistributeClocks":
        # value: True = multiple crates
        assert isinstance(
            value, bool
        ), "DistributeClocks needs to be either True or False"
        bit_list[31 - 11] = value
    elif name == "SortEvents":
        # value: True = sort events based on timestamp
        assert isinstance(value, bool), "SortEvents needs to be either True or False"
        bit_list[31 - 12] = value
    elif name == "ConnectFastTriggerBP":
        # value: True = Connect the fast trigger to the backplane
        assert isinstance(
            value, bool
        ), "ConnectFastTriggerBP needs to be either True or False"
        bit_list[31 - 13] = value

    # Channel parameters  (currently CHAN_CSRA)
    elif name == "FastTrigSelect":
        # value: 'external' or 'group'
        assert value in [
            "external",
            "group",
        ], "FastTrigSelect needs to be either 'external' or 'group'"
        bit = value == "external"
        bit_list[31 - 0] = bit
    elif name == "ModValSignal":
        # value: 'modgate' or 'global'
        assert value in [
            "modgate",
            "global",
        ], "ModValSignal needs to be either 'modgate' or 'global'"
        bit = value == "modgate"
        bit_list[31 - 1] = bit
    elif name == "GoodChannel":
        # value: True = enable channel
        assert isinstance(value, bool), "GoodChannel needs to be either True or False"
        bit_list[31 - 2] = value
    elif name == "ChanValSignal":
        # value: 'channelgate' or 'channelvalidation'
        assert value in [
            "channelgate",
            "channelvalidation",
        ], "ChanValSignal needs to be either 'channelgate' or 'channelvalidation'"
        bit = value == "channelgate"
        bit_list[31 - 3] = bit
    elif name == "RejectIfFull":
        # value: True = reject data if buffer is full
        assert isinstance(value, bool), "RejectIfFull needs to be either True or False"
        bit_list[31 - 4] = value
    elif name == "Polarity":
        # value: True=positive slope, False=negative slope
        assert value in [
            "positive",
            "negative",
        ], "Polarity needs to be either 'positive' or 'negative'"
        bit = value == "positive"
        bit_list[31 - 5] = bit
    elif name == "EnableVeto":
        # value: True = enable veto
        assert isinstance(value, bool), "EnableVeto needs to be either True or False"
        bit_list[31 - 6] = value
    elif name == "CaptureHistogram":
        # value: True = enable capture of MCA histograms
        assert isinstance(
            value, bool
        ), "CaptureHistogram needs to be either True or False"
        bit_list[31 - 7] = value
    elif name == "CaptureTrace":
        # value: True = enable capture trace
        assert isinstance(value, bool), "CaptureTrace needs to be either True or False"
        bit_list[31 - 8] = value
    elif name == "EnableQDC":
        # value: True = enable capture QDC sums
        assert isinstance(value, bool), "EnableQDC needs to be either True or False"
        bit_list[31 - 9] = value
    elif name == "EnableCFD":
        # value: True = enable CFD
        assert isinstance(value, bool), "EnableCFD needs to be either True or False"
        bit_list[31 - 10] = value
    elif name == "EnableModVal":
        # value: True = enable module validation
        assert isinstance(value, bool), "EnableModVal needs to be either True or False"
        bit_list[31 - 11] = value
    elif name == "CaptureSums":
        # value: True = enable capture raw energy susms
        assert isinstance(value, bool), "CaptureSums needs to be either True or False"
        bit_list[31 - 12] = value
    elif name == "EnableChannelVal":
        # value: True = enable channel validation
        assert isinstance(
            value, bool
        ), "EnableChannelVal needs to be either True or False"
        bit_list[31 - 13] = value
    elif name == "Gain":
        # value: 2.5 (True) or 0.625 (False)
        assert value in [2.5, 0.625], "Gain needs to be either 0.625 or 2.5"
        bit = value == 2.5
        bit_list[31 - 14] = bit
    elif name == "RejectPileup":
        # value: 'all' (no energies for pileup events),
        #        'single' (reject pileup),
        #        'pileup' (trace, timestamp for pileup, no trace for single)
        #        'pileup-only' (only record trace, timestamp, etc for pileup
        #                       events, no single events)
        assert value in [
            "single",
            "pileup-only",
            "pileup",
            "all",
        ], "RejectPileup needs to be either 'all', 'single', 'pileup', 'pileup-only'"
        bit0 = (value == "single") or (value == "pileup-only")
        bit1 = (value == "pileup") or (value == "pileup-only")
        bit_list[31 - 15] = bit0
        bit_list[31 - 16] = bit1
    elif name == "SkipLargePulses":
        # value: True = don't record traces for large pulses
        assert isinstance(
            value, bool
        ), "SkipLargePulses needs to be either True or False"
        bit_list[31 - 17] = value
    elif name == "GroupTrigSignal":
        # value: 'external' or 'local'
        assert value in [
            "external",
            "local",
        ], "GroupTrigSignal needs to be either 'external' or 'local'"
        bit = value == "external"
        bit_list[31 - 18] = bit
    elif name == "ChannelVetoSignal":
        # value: 'channel' or 'front'
        assert value in [
            "channel",
            "front",
        ], "ChannelVetoSignal needs to be either 'channel' or 'front'"
        bit = value == "channel"
        bit_list[31 - 19] = bit
    elif name == "ModVetoSignal":
        # value: 'module' or 'front'
        assert value in [
            "module",
            "front",
        ], "ModVetoSignal needs to be either 'module' or 'front'"
        bit = value == "module"
        bit_list[31 - 20] = bit
    elif name == "ExtTimestamps":
        # value: True = include external timestamps in header
        assert isinstance(value, bool), "ExtTimestamps needs to be either True or False"
        bit_list[31 - 21] = value
    return bit_list


def change_raw_setting_from_dict(settings: Dict, call_pixie: Optional[bool] = True):
    """Modify the raw settings on the pixie.

    Note: this bypasses any checks that are normally done in XIA's C-libary. This can be
    used to, for example, modify `Peaksample` if needed. However, if you set parameters
    to invalid numbers, there will be no error and the binary stream can be invalid.

    Parameters
    ----------
    settings
        A dictionary with the settings names as key and the values for each channel
        (or a single value that will be applied to all channels). One key must be called
        `channels` and must be a list of (module, channel) tuples.
    call_pixie
        Can be set to False, in which case the pixie16 is not actually called (used for tests)
    """
    assert (
        "channels" in settings
    ), "The settings dictionary needs an entry listing called 'channels' the channels"
    channels = settings.pop("channels")

    for c in channels:
        assert isinstance(
            c, (list, tuple)
        ), "Setting dictionary: each channel must be a list or tuple"
        assert (
            len(c) == 2
        ), "Setting dictionary: each channel must have two entries (modules, channel)"

    # get all the settings data from the pixie
    if call_pixie:
        current = read_settings()
    else:
        current = read_dummy_settings(
            list({mod for (mod, ch) in channels})
        )  # mostly used for testing

    # update all settings
    for name, value in settings.items():
        for i, (module, channel) in enumerate(channels):
            current_value = current[module].get_by_name(name)

            if isinstance(current_value, (bool, int, float)):
                assert isinstance(value, (bool, int, float))
                current[module].set_by_name(name, value)
            if isinstance(current_value, (list, tuple)):
                if isinstance(value, (bool, int, float)):
                    value_list = [value] * len(channels)
                else:
                    value_list = list(value)
                current_value[channel] = value_list[i]
                current[module].set_by_name(name, current_value)

    if call_pixie:
        write_settings(current)
    return current


def set_defaults_for_module(module: int, call_pixie: Optional[bool] = True):
    """Sets defaults for all used module and channel parameters in a module.

    Parameters
    ----------
    module
        The module number (0, 1, ...)
    call_pixie
        Can be set to False, in which case the pixie16 is not actually called (used for tests)

    """
    channels = [(module, x) for x in range(16)]
    settings = {
        "channels": channels,
        "BASELINE_AVERAGE": 3,
        "BASELINE_PERCENT": 10,
        "BINFACTOR": 1,
        "BLCUT": 200,
        "CaptureHistogram": False,
        "CaptureSums": False,
        "CaptureTrace": False,
        "CFDDelay": 8,
        "CFDScale": 1,
        "CFDThresh": 100,
        "CHANNEL_CSRA": 0,
        "ChanTrigStretch": 0.01,
        "ChanValSignal": "channelvalidation",
        "ChannelVetoSignal": "front",
        "ModVetoSignal": "front",
        "ExtTimestamps": False,
        "EMIN": 100,
        "EnableCFD": False,
        "EnableChannelVal": False,
        "ENERGY_FLATTOP": 0.04,  # as above
        "ENERGY_RISETIME": 0.020,  # times 2**SLOW_FILTER_RANGE
        "ExternDelayLen": 0.01,
        "ExtTrigStretch": 0.01,
        "FAST_FILTER_RANGE": 0,
        "FASTTRIGBACKLEN": 0.01,
        "FastTrigBackplaneEna": 0,
        "FastTrigSelect": "group",
        "FtrigoutDelay": 0.01,
        "Gain": 0.625,
        "GoodChannel": True,
        "GroupTrigSignal": "local",
        "IN_SYNCH": 0,
        "MODULE_CSRB": 0,
        "MultiplicityMaskH": 0,
        "MultiplicityMaskL": 0,
        "Polarity": "negative",
        "QDCLen0": 0.06,
        "QDCLen1": 0.06,
        "QDCLen2": 0.06,
        "QDCLen3": 0.06,
        "QDCLen4": 0.06,
        "QDCLen5": 0.06,
        "QDCLen6": 0.06,
        "QDCLen7": 0.06,
        "RejectIfFull": True,
        "RejectPileup": "all",
        "SkipLargePulses": False,
        "SLOW_FILTER_RANGE": 1,
        "SYNCH_WAIT": 0,
        "TAU": 0.03,
        "TRACE_DELAY": 0.4,
        "TRACE_LENGTH": 1.0,
        "TrigConfig0": 0,
        "TrigConfig1": 0,
        "TrigConfig2": 0,
        "TrigConfig3": 0,
        "TRIGGER_FLATTOP": 0.02,
        "TRIGGER_RISETIME": 0.02,
        "TRIGGER_THRESHOLD": 1_000,
        "VetoStretch": 0.30,
        "VOFFSET": 0.0,
        "XDT": 0.1,
    }
    return change_setting_from_dict(settings, call_pixie=call_pixie)



import numpy as np

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QMainWindow

from ..plot import PlotOnly


class PlotCoincidenceTraces(PlotOnly):
    """Create the plot of fast triggers, delayed triggers, coincidences, validations, etc.

    Devides the axes in several plotting areas: Fast triggers, coincidences, validation.

    Each area is divides by `large_gap` and the plots in each area are divided by `gap`.

    The class provides functions to make plotting easier.

    """

    gap = 0.2
    large_gap = 1.6

    def __init__(self, main: QMainWindow, N, maximum):
        super().__init__(main)
        self.N = N

        self.maximum = maximum

        self.cursor_select = False
        self.fig.canvas.mpl_connect("motion_notify_event", self.onmove)
        # need the following to be able to listen to key events
        self.fig.canvas.mpl_connect(
            "axes_enter_event", lambda x: self.fig.canvas.setFocus()
        )

        self.calc_positions()

    def calc_positions(self):
        height_validation = 3 * (1 + self.gap) * self.N - self.gap
        height_coincidence = (1 + self.gap) * self.N
        height_fast_trigger = 2 * (1 + self.gap) * self.N - self.gap

        self.position_validation = height_validation
        self.position_coincidence = (
            self.position_validation + self.large_gap + height_coincidence
        )
        self.position_fast_trigger = (
            self.position_coincidence + self.large_gap + height_fast_trigger
        )

        self.FT_yrange = [[0, 1]] * self.N  # we set the correct values while plotting
        self.x = np.arange(0, self.maximum, 1)

    def get_y_range(self, y):
        for i, [down, up] in enumerate(self.FT_yrange):
            if down < y < up:
                return i
        return

    def in_y_range(self, y):
        for down, up in self.FT_yrange:
            if down < y < up:
                return True
        return False

    def in_x_range(self, x):
        return 0 < x < self.maximum

    def onmove(self, event):
        y = event.ydata
        if y is None:
            if self.cursor_select:
                QApplication.restoreOverrideCursor()
                self.cursor_select = False
            return

        x = int(event.xdata)

        if self.in_x_range(x) and self.in_y_range(y):
            if not self.cursor_select:
                QApplication.setOverrideCursor(Qt.CrossCursor)
                self.cursor_select = True
        else:
            if self.cursor_select:
                QApplication.restoreOverrideCursor()
                self.cursor_select = False

    def plot_titles(self):
        self.text(-40, self.position_fast_trigger + 1.2, "Fast Triggers and Delays")
        self.text(-40, self.position_coincidence + 1.2, "Coincidence Trigger")
        self.text(-40, self.position_validation + 1.2, "Validation")

    def plot_fast_trigger_and_coinc_window(self, ft, window, i, channel):
        position = self.position_fast_trigger - i * 2 * (1 + self.gap)

        self.FT_yrange[i] = [position, position + 1]

        self.plot(self.x, ft + position, color="black")
        self.text(-30, position, f"Ch {channel}")
        position -= self.gap + 1
        self.plot(self.x, window + position, color="orange")

    def plot_coincidence(self, y, i, channel):
        position = self.position_coincidence - i * (1 + self.gap)

        self.plot(self.x, y + position, color="blue")
        self.text(-30, position, f"Ch {channel}")

    def plot_validation(self, coinc, validation, ft, delayed, i, channel):
        position = self.position_validation - i * 3 * (1 + self.gap)

        self.plot(self.x, coinc + position, color="blue")
        position -= self.gap + 1
        self.plot(self.x, validation + position, color="green")
        validated = np.sum(delayed * validation)
        if validated:
            color = "green"
        else:
            color = "red"
        self.text(-30, position, f"Ch {channel}", backgroundcolor=color)
        position -= self.gap + 1
        self.plot(self.x, ft + position, color="black")
        self.plot(self.x, delayed + position, color="red")



"""Functions that allow a 1d and 2d parameter scan using the pixie16.

During the parameter scan, one or two settings are changed and a MCA
histogram. The histogram is integrated and plotted against the
parameter.

"""
from dataclasses import dataclass
import sys
import time
from typing import Optional

from matplotlib import pyplot as plt
import numpy as np
from rich.progress import Progress

from . import control

PARAMETERS = [
    ["TRIGGER_RISETIME", "μs", "FASTLENGTH"],
    ["TRIGGER_FLATTOP", "μs", "FASTGAP"],
    ["TRIGGER_THRESHOLD", "ADC units", "FASTTHRESH"],
    ["ENERGY_RISETIME", "μs", "SLOWLENGTH"],
    ["ENERGY_FLATTOP", "μs", "SLOWGAP"],
    ["TAU", "μs", "PREAMPTAU"],
    ["TRACE_LENGTH", "μs", "TRACELENGTH"],
    ["TRACE_DELAY", "μs", "TRIGGERDELAY"],
    ["VOFFSET", "V", "OFFSETDAC"],
    ["XDT", "μs", "XWAIT"],
    ["BASELINE_PERCENT", "%", "BASELINEPERCENT"],
    ["EMIN", "None", "ENERGYLOW"],
    ["BINFACTOR", "None", "LOG2EBIN"],
    ["CHANNEL_CSRA", "bit pattern", "CHANCSRA"],
    ["CHANNEL_CSRB", "bit pattern", "CHANCSRB"],
    ["BLCUT", "None", "BLCUT"],
    ["FASTTRIGBACKLEN", "μs", "FASTTRIGBACKLEN"],
    ["CFDDelay", "μs", "CFDDELAY"],
    ["CFDScale", "None", "CFDSCALE"],
    ["QDCLen0", "μs", "QDCLEN0"],
    ["QDCLen1", "μs", "QDCLEN1"],
    ["QDCLen2", "μs", "QDCLEN2"],
    ["QDCLen3", "μs", "QDCLEN3"],
    ["QDCLen4", "μs", "QDCLEN4"],
    ["QDCLen5", "μs", "QDCLEN5"],
    ["QDCLen6", "μs", "QDCLEN6"],
    ["QDCLen7", "μs", "QDCLEN7"],
    ["ExtTrigStretch", "μs", "EXTTRIGSTRETCH"],
    ["MultiplicityMaskL", "bit pattern", "MULTIPLICITYMASKL"],
    ["MultiplicityMaskH", "bit pattern", "MULTIPLICITYMASKH"],
    ["ExternDelayLen", "μs", "EXTERNDELAYLEN"],
    ["FtrigoutDelay", "μs", "FTRIGOUTDELAY"],
    ["ChanTrigStretch", "μs", "?"],
    ["VetoStretch", "μs", "?"],
]
PARAMETER_NAMES = [x[0] for x in PARAMETERS]


@dataclass
class ScanParameter:
    """A convenient class to store scan parameters."""

    name: str
    module: int
    channel: int
    start: float
    stop: float
    step: float

    def range(self) -> np.ndarray:
        """Return the values that should be scanned."""
        return np.linspace(
            self.start, self.stop, (self.stop - self.start) / self.step + 1
        )


def parameter_scan_1d(
    parameter: ScanParameter,
    dwelltime: int = 0,
    title: Optional[str] = None,
    savefile: Optional[str] = None,
):
    """Do a 1D parameter scan on the pixie16.

    This function assumes that the pixie16 is already booted and ready to take MCA data.
    """
    if parameter.name not in PARAMETER_NAMES:
        print(
            f"ERROR: parameter {parameter.name} not known. Needs ot be one of {','.join(PARAMETER_NAMES)}"
        )
        sys.exit(1)

    if title is None:
        title = "1d-Scan"

    myrange = parameter.range()

    X = []
    Y = []
    fig, ax = plt.subplots()

    with Progress() as progress:
        task = progress.add_task(total=len(myrange))
        for x in myrange:
            progress.update(task, description=f"Value: {x:.1f} ->")
            value = control.set_channel_parameter(
                parameter.name, x * 1e-3, parameter.module, parameter.channel
            )
            X.append(value)

            time.sleep(dwelltime)

            control.start_histogram_run()

            check = 1
            while check == 1:
                # Check Run Status
                plt.pause(1)
                check = control.check_run_status()
            out = control.read_histograms((parameter.module, parameter.channel))
            Y.append(np.sum(out))

            ax.clear()
            ax.set_title(title)
            ax.plot(X, Y, ".-")
            ax.set_xlabel(
                f"M{parameter.module}-ch{parameter.channel} {parameter.name} [ns]"
            )
            ax.set_ylabel("counts")
            fig.canvas.draw()
            plt.pause(0.001)
            progress.advance(task)

    if savefile:
        plt.savefig(savefile)
        print(f"saved plot in {savefile}")

    plt.show(block=True)
    return X, Y


def parameter_scan_2d(
    parameterA: ScanParameter,
    parameterB: ScanParameter,
    dwelltime: int = 0,
    title: Optional[str] = None,
    savefile: Optional[str] = None,
):
    """Do a 2D parameter scan on the pixie16.

    This function assumes that the pixie16 is already booted and ready to take MCA data.

    The data plotted will be from moduleB/channelB.
    """
    for parameter in [parameterA.name, parameterB.name]:
        if parameter not in PARAMETER_NAMES:
            print(
                f"ERROR: parameter {parameter.name} not known. Needs ot be one of {','.join(PARAMETER_NAMES)}"
            )
            sys.exit(1)

    if title is None:
        title = "2d-Scan"

    myrangeA = parameterA.range()
    myrangeB = parameterB.range()

    X = []
    Y = []
    Z = np.zeros((len(myrangeA), len(myrangeB)))

    fig, ax = plt.subplots()

    with Progress() as progress:
        task = progress.add_task(total=len(myrangeA) * len(myrangeB))
        for i, x in enumerate(myrangeA):
            Xvalue = control.set_channel_parameter(
                parameterA.name, x * 1e-3, parameterA.module, parameterA.channel
            )
            for j, y in enumerate(myrangeB):
                progress.update(description=f"X={x:.1f} Y={y:.1f} ->")
                Yvalue = control.set_channel_parameter(
                    parameterB.name, y * 1e-3, parameterB.module, parameterB.channel
                )
                X.append(Xvalue)
                Y.append(Yvalue)

                time.sleep(dwelltime)

                control.start_histogram_run()

                check = 1
                while check == 1:
                    # Check Run Status
                    plt.pause(1)
                    check = control.check_run_status()
                out = control.read_histograms((parameterB.module, parameterB.channel))
                Z[i, j] = np.sum(out)

                ax.clear()
                ax.set_title(title)
                ax.imshow(
                    Z.T,
                    origin="lower",
                    extent=[myrangeA[0], myrangeA[-1], myrangeB[0], myrangeB[-1]],
                )
                ax.set_xlabel(
                    f"M{parameterA.module}-ch{parameterA.channel} {parameterA.name} [ns]"
                )
                ax.set_ylabel(
                    f"M{parameterB.module}-ch{parameterB.channel} {parameterB.name} [ns]"
                )
                fig.canvas.draw()
                plt.pause(0.001)
                progress.advance(task)

    if savefile:
        plt.savefig(savefile)
        print(f"saved plot in {savefile}")

    plt.show(block=True)
    return X, Y, Z



import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from PyQt5.QtWidgets import (
    QMainWindow,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class MyNavigationToolbar(NavigationToolbar):
    """Only display the buttons we need"""

    toolitems = [
        t
        for t in NavigationToolbar.toolitems
        if t[0] in ("Home", "Back", "Forward", "Pan", "Zoom", "Save")
    ]


class Plot(QWidget):
    """A Widget with a matplotlib figure and and (optional) navigation toolbar.

    We already create a figure and single axis to plot on.

    """

    def __init__(self, main: QMainWindow, with_navbar: bool = True) -> None:
        super().__init__()
        self.main = main
        self.update = True  # do we need to update when switching tabs?

        self.fig = plt.figure()
        self.axes = self.fig.subplots()
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.vbox = QVBoxLayout()

        if with_navbar:
            self.nav = MyNavigationToolbar(self.canvas, self, coordinates=False)
            self.nav.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            self.vbox.addWidget(self.nav)

        self.vbox.addWidget(self.canvas)

        self.setLayout(self.vbox)


class PlotOnly(Plot):
    """A matplotlib area with a single axes and without borders and navbar.

    Also added shortcuts to clear the plot, add text, and plot data using steps.
    """

    def __init__(self, main: QMainWindow):
        super().__init__(main, with_navbar=False)

        self.axes.axis("off")

    def clear(self):
        self.axes.clear()
        self.axes.axis("off")

    def plot(self, x, y, **kwargs):
        kwargs["drawstyle"] = "steps-mid"
        self.axes.plot(x, y, **kwargs)

    def text(self, x, y, text, **kwargs):
        self.axes.text(x, y, text, **kwargs)



import numpy as np
import struct
import json
from pathlib import Path
from typing import Union, List, Optional

from .units import SETTINGS_UNITS_TBL
from .file_format import SETTINGS_FILE_FORMAT
from .field_classes import (
    Entity,
    ABField,
    BinaryField,
    OffsetVoltageField,
    TraceDelayField,
)


def combine_hi_lo_registers(hi, lo):
    """Creates a whole parameter that was split into two 32-bit parameters A and B."""
    return (np.array(hi) * 2 ** 32 + np.array(lo)).astype(float)


def _create_settings_ureg(fast_filter_range, slow_filter_range):
    """Creates a pint unit registry with pixie units specific to the settings of this run."""
    import pint

    ureg = pint.UnitRegistry()
    ureg.define("ADC_cycles = 2*ns")
    ureg.define("FPGA_cycles = 5*ADC_cycles")
    ureg.define(f"fast_filter_cycles = FPGA_cycles*2**{fast_filter_range}")
    ureg.define(f"slow_filter_cycles = FPGA_cycles*2**{slow_filter_range}")
    return ureg


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types."""

    def default(self, obj):
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class OldSettings(Entity):
    """Class that holds settings data from the pixie16. All settings values can be accessed as attributes or through
    the get method."""

    # Derived fields. Look at field_classes.py for implementations/field logic
    LiveTime = ABField()
    FastPeaks = ABField()
    RealTime = ABField()
    RunTime = ABField()
    MultiplicityMaskL = BinaryField()
    MultiplicityMaskH = BinaryField()
    FastTrigBackplaneEna = BinaryField()
    TrigConfig = BinaryField()
    ChanCSRa = BinaryField()
    ChanCSRb = BinaryField()
    ModCSRA = BinaryField()
    ModCSRB = BinaryField()
    OffsetVoltage = OffsetVoltageField()
    TraceDelay = TraceDelayField()

    def __init__(self, **kwargs):
        # set all key value pairs as attributes
        self.__dict__.update(kwargs)
        # create a unit registry with pixie units specific to the settings of this run
        self.ureg = _create_settings_ureg(self.FastFilterRange, self.SlowFilterRange)

    def __getitem__(self, item):
        return self.get(item)

    def get(self, key, channel=None, as_pint=False):
        """A utility function to get a channel's settings value or get a settings value with units."""
        value = getattr(self, key)
        if channel is not None:
            value = value[channel]
        if as_pint:
            units = getattr(self.ureg, SETTINGS_UNITS_TBL[key])
            value = value * units
        return value

    @classmethod
    def _from_file(cls, file: str, module: int = 0):
        """Create a settings dictioanry from .set file for a given module."""
        # get bytes from file
        with open(file, "rb") as fp:
            binary_data = fp.read()
        MODULE_SIZE = 4 * 1280
        binary_data = binary_data[module * MODULE_SIZE : (module + 1) * MODULE_SIZE]

        # unpack field values from bytes
        settings_dict = {}
        for field, (offset, fmt) in SETTINGS_FILE_FORMAT.items():
            settings_dict[field] = struct.unpack_from(fmt, binary_data, offset)

        # unpack singletons
        for key, value in settings_dict.items():
            if len(value) == 1:
                settings_dict[key] = value[0]
            else:
                settings_dict[key] = value

        return settings_dict


def convert_old_settings(
    settings_filename: Union[str, Path],
    modules: List = [0],
    metadata_filename: Optional[Union[str, Path]] = None,
):
    """Convert an old .set file to a .json file.

    .set files do not include the metadata (mostly hardware) information included
    in the .json settings file. If one provides a path (via the optional
    metadata_filename argument) to a new .json settings file that already has
    metadata, the metadata will be added to the .json converted from the .set file."""

    channel_input_params = [
        "BLcut",
        "BaselinePercent",
        "CFDDelay",
        "CFDScale",
        "CFDThresh",
        "ChanCSRa",
        "ChanCSRb",
        "ChanTrigStretch",
        "DigGain",
        "EnergyLow",
        "ExtTrigStretch",
        "ExternDelayLen",
        "FastGap",
        "FastLength",
        "FastThresh",
        "FastTrigBackLen",
        "FtrigoutDelay",
        "GainDAC",
        "Integrator",
        "Log2Bweight",
        "Log2Ebin",
        "MultiplicityMaskH",
        "MultiplicityMaskL",
        "OffsetDAC",
        "PAFlength",
        "PSAlength",
        "PSAoffset",
        "PeakSample",
        "PeakSep",
        "PreampTau",
        "QDCLen0",
        "QDCLen1",
        "QDCLen2",
        "QDCLen3",
        "QDCLen4",
        "QDCLen5",
        "QDCLen6",
        "QDCLen7",
        "ResetDelay",
        "SlowGap",
        "SlowLength",
        "ThreshWidth",
        "TraceLength",
        "TrigOutLen",
        "TriggerDelay",
        "VetoStretch",
        "Xavg",
        "Xwait",
    ]

    module_input_params = [
        "ChanNum",
        "CoincPattern",
        "CoincWait",
        "ControlTask",
        "CrateID",
        "FIFOLength",
        "FastFilterRange",
        "FastTrigBackplaneEna",
        "HostIO",
        "HostRunTimePreset",
        "InSynch",
        "MaxEvents",
        "ModCSRA",
        "ModCSRB",
        "ModFormat",
        "ModID",
        "ModNum",
        "Resume",
        "RunTask",
        "SlotID",
        "SlowFilterRange",
        "SynchWait",
        "TrigConfig",
        "U00",
        "UserIn",
    ]

    stats_params = [
        "RealTimeA",
        "RealTimeB",
        "LiveTimeA",
        "LiveTimeB",
        "FastPeaksA",
        "FastPeaksB",
        "ChanEventsA",
        "ChanEventsB",
    ]

    settings_filename = str(settings_filename)
    # Make sure filename extension is .set
    if not settings_filename.endswith(".set"):
        if settings_filename.endswith(".json"):
            print("[INFO] Settings file is already a json.")
        else:
            print("[ERROR] Settings file is neither a .set nor a .json.")
        print("Not converting settings.")
        return

    new_settings_lst = []
    new_stats_lst = []
    new_settings_filename = settings_filename[:-3] + "json"
    new_stats_filename = new_settings_filename.replace("settings", "stats")

    for module in modules:
        # Load settings and prepare to rewrite settings and stats to dictionaries
        old_settings_dict = OldSettings._from_file(settings_filename, module)
        new_settings_dict = {
            "channel": {"input": {}},
            "metadata": "None",
            "module": {"input": {}},
        }
        new_stats_dict = {"module": module}
        if metadata_filename:
            if not metadata_filename.endswith(".json"):
                print("[ERROR] Settings file with metadata must be a .json file.")
                print("Not writing metadata to settings file.")
            else:
                with open(metadata_filename) as file:
                    metadata = json.load(file)[0]["metadata"]
                new_settings_dict["metadata"] = metadata

        # Organize settings from old file into dictionaries in the new settings/stats format
        for key, value in sorted(old_settings_dict.items()):
            # Channel Parameters
            if key in channel_input_params:
                new_settings_dict["channel"]["input"][key] = value
            # Module Parameters
            elif key in module_input_params:
                new_settings_dict["module"]["input"][key] = value
            # Stats Parameters
            elif key in stats_params:
                # If a parameter has only one value, apply it to all 16 channels
                if isinstance(value, (int, float)):
                    value = [value] * 16
                if key.endswith("A"):
                    A_value = value
                elif key.endswith("B"):
                    combined_key = key[:-1]
                    # Combine A and B 32-bit parameters
                    combined_val = combine_hi_lo_registers(A_value, value)
                    if combined_key in ["RealTime", "LiveTime"]:
                        # Convert from clock cycles to ns (multiply by 10)
                        # Convert from ns to s (divide by 1e9)
                        combined_val = combined_val / 1e8
                        if combined_key == "RealTime":
                            combined_key = "real_time"
                        elif combined_key == "LiveTime":
                            combined_key = "live_time"
                    elif combined_key == "FastPeaks":
                        combined_key = "input_counts"
                    elif combined_key == "ChanEvents":
                        combined_key = "output_counts"
                    new_stats_dict[combined_key] = combined_val
        # Add module settings dictionary to a list of settings dictionaries
        new_settings_lst.append(new_settings_dict)

        # Add input and output count rate to stats dictionary
        new_stats_dict["input_count_rate"] = (
            new_stats_dict["input_counts"] / new_stats_dict["live_time"]
        )
        new_stats_dict["output_count_rate"] = (
            new_stats_dict["output_counts"] / new_stats_dict["real_time"]
        )
        # Assumes dictionaries are ordered, which is only for python >= 3.6
        key_order = [
            "module",
            "real_time",
            "live_time",
            "input_counts",
            "input_count_rate",
            "output_counts",
            "output_count_rate",
        ]
        new_stats_dict_ordered = {}
        # Convert values in stats dictionary from numpy arrays to lists
        for key in key_order:
            if key != "module":
                new_stats_dict[key] = new_stats_dict[key].tolist()
            new_stats_dict_ordered[key] = new_stats_dict[key]
        # Add module stats dictionary to a list of stats dictionaries
        new_stats_lst.append(new_stats_dict_ordered)

    # Write new settings .json file
    new_settings_json = json.dumps(
        new_settings_lst,
        indent=4,
        separators=(",", ": "),
        sort_keys=True,
        cls=NumpyEncoder,
    )
    with open(new_settings_filename, "a") as f:
        f.write(new_settings_json + "\n")

    # Write new stats .json file
    new_stats_json = json.dumps(new_stats_lst)
    with open(new_stats_filename, "a") as f:
        f.write(new_stats_json + "\n")

    return new_settings_filename, new_stats_filename



"""Some helper functions."""

from contextlib import contextmanager

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCursor


@contextmanager
def busy_cursor(app):
    """Swap out the cursor for a spinning wheel or similar."""
    app.setOverrideCursor(QCursor(Qt.WaitCursor))
    try:
        yield
    finally:
        app.restoreOverrideCursor()



"""The main implementation of the Settings class."""

from collections import UserDict, UserList
import json
from pathlib import Path
import struct
from typing import List, Dict, Union

from .units import SETTINGS_UNITS_TBL
from .file_format import SETTINGS_FILE_FORMAT
from .field_classes import (
    Entity,
    ABField,
    BinaryField,
    OffsetVoltageField,
    TraceDelayField,
)
from .old_settings_class import convert_old_settings


def load_settings(file, module=0):
    """Takes the name of a settings file and the index of the module of interest. Returns a Settings object."""
    return Settings.from_file(file, module)


def load_stats(file, module=0):
    """Takes the name of a settings file and the index of the module of interest. Returns a Settings object."""
    return Stats.from_file(file, module)


class SettingsBase(UserDict):
    """A 1:1 mapping of the settings file to a dictionary.

    We also provide some helper methods to more easily access the values.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = {}
        self.duplicate_keys = []

    def create_index(self, d: Dict = None, level: List = None):
        """Create an index.

        We use the index later to be able to look up

        self["channel"]["input"]["VetoStretch"]
        by just using
        self.get_by_name("VetoStretch")

        At the same time, we keep track of duplicate keys, so that we
        can raise a KeyError for those.
        """

        if d is None:
            d = self
        if level is None:
            level = []

        keys = {}

        for k, v in d.items():
            if isinstance(v, dict):
                new_level = level.copy()
                new_level.append(k)
                new_keys = self.create_index(v, new_level)
                for nk, nv in new_keys.items():
                    if nk in self:
                        self.duplicate_keys.append(nk)
                    keys[nk] = nv
            else:
                keys[k] = level.copy()

        if d == self:
            self.index = keys

        return keys

    @classmethod
    def from_file(cls, file: Union[Path, str], module: int = 0):
        """Creates a Settings object from a file."""

        if isinstance(file, str):
            file = Path(file)

        if file.suffix != ".json":
            if file.suffix == ".set":
                print("[INFO]: Converting settings file from .set to .json")
                file = convert_old_settings(file, [0])
            else:
                print("[ERROR]: Unkonwn file type for setting file (shoudl be .json)")
            return

        with file.open("r") as fp:
            data = json.load(fp)
        setting = cls()
        setting.data = data[module]
        setting.create_index()
        return setting

    @classmethod
    def from_dict(cls, data: Dict):
        """Creates a Settings object from a file."""
        setting = cls()
        setting.data = data
        setting.create_index()
        return setting

    def get_by_name(self, name: str):
        """Alternative way to access keys in the dictionary.

        The advantage of using `get` is that it will look at all
        levels inside the setting dict and return the value. This only
        works if the key is unique. Otherwise, we raise an exception.
        """
        if name in self.duplicate_keys:
            raise KeyError

        level = self.index[name]
        d = self
        for l in level:
            d = d[l]
        return d[name]

    def set_by_name(self, name: str, value) -> None:
        """Alternative way to access keys in the dictionary.

        The advantage of using `set` is that it will look at all
        levels inside the setting dict and set the correct value. This only
        works if the key is unique. Otherwise, we raise an exception.
        """
        if name in self.duplicate_keys:
            raise KeyError

        level = self.index[name]
        d = self
        for l in level:
            d = d[l]
        d[name] = value


class Settings(SettingsBase):
    pass


class Stats(SettingsBase):
    pass


class CrateSettings(UserList):
    def __init__(self, input=None):
        if input is None:
            self.data = []
        else:
            self.data = list(input)
            for x in self.data:
                assert isinstance(
                    x, Settings
                ), "Can only put Seetings in CrateSettings."



"""Interface to C-library function provided by XIA.

The interface to python is provided by the ctypes library.

Currently calling the XIA C-library functions is only supported on
Windows. Adding Linux support should be relatively straight forward
though (just changing the ctype loading of the library).

We provide access to all XIA functions and try to provide a
more pythonic interface. This include providing our own names for the
functions (mostly just removing the Pixie16 in front and converting
from CamelCase to lower case seperated by underscores).

This file only includes the functions that are in XIA's library and a
very few helper functions. We see these as more low level functions
and build our own set of python function on top of these in
`control.py` that automate several aspects, e.g. providing information
from a config file, etc. See config.py for more details. There should
be no need for the user to call any of the functions in here directly.

XIA's code can be found here:
  https://github.com/xiallc/pixie_sdk
Their document can be accessed here:
  https://docs.pixie16.xia.com/

This is based on version 3.3a2 of XIA's SDK.

Here is the list of supported functions and their names in our package:

.. list-table:: Functions
   :widths: 25 25
   :header-rows: 1

   * - SDK name
     - python name
   * - Pixie16AcquireADCTrace, Pixie16ReadSglChanADCTrace
     - acquire_adc_trace
   * - Pixie16AdjustOffsets
     - adjust_offsets
   * - Pixie16BLcutFinder
     - baseline_cut_finder
   * - Pixie16BootModule
     - boot_module
   * - Pixie16CheckExternalFIFOStatus
     - check_external_FIFO_status
   * - Pixie16CheckRunStatus
     - check_run_status
   * - Pixi16CopyDSPParameters
     - copy_DSP_parameters
   * - Pixie16EndRun
     - end_run
   * - Pixie16ExitSystem
     - exit_system
   * - Pixie16InitSystem
     - init_system
   * - Pixie16LoadDSPParametersFromFile
     - load_dsp_parameters_from_file
   * - Pixie16ReadDataFromExternalFIFO
     - read_data_from_external_FIFO
   * - Pixie16ReadHistogramFromModule
     - read_histogram_from_module
   * - Pixie16ReadModuleInfo
     - read_module_info
   * - Pixie16ReadSglChanBaselines, Pixie16AcquireBaselines
     - read_baselines
   * - Pixie16ReadSglChanPar
     - read_single_channel_parameter
   * - Pixie16ReadSglModPar
     - read_single_module_parameter
   * - Pixie16ReadStatisticsFromModule,
       Pixie16GetStatisticsSize,
       Pixie16ComputeRealTime,
       Pixie16ComputeLiveTime,
       Pixie16ComputeRawInputCount,
       Pixie16ComputeInputCountRate,
       Pixie16ComputeRawOutputCount,
       Pixie16ComputeOutputCountRate
     - read_statistics_from_module
   * - Pixie16SaveDSPParametersToFile
     - save_dsp_parameters_as_json
   * - Pixie16SetDACs
     - set_DACs
   * - Pixie16StartHistogramRun
     - start_histogram_run
   * - Pixie16StartListModeRun
     - start_list_mode_run
   * - Pixie16TauFinder
     - tau_finder
   * - Pixie16WriteSglChanPar
     - write_single_channel_parameter
   * - Pixie16WriteSglModPar
     - write_single_module_parameter

XIA's library provides several other functions that we do not expose
direclty, but use in some of our calls. For example,
`acquire_adc_trace` include the XIA call to set up the measurement.

"""

__all__ = [
    "acquire_adc_trace",
    "adjust_offsets",
    "baseline_cut_finder",
    "boot_module",
    "check_external_FIFO_status",
    "check_run_status",
    "end_run",
    "exit_system",
    "init_system",
    "load_dsp_parameters_from_file",
    "read_data_from_external_FIFO",
    "read_histogram_from_module",
    "read_module_info",
    "read_baselines",
    "read_single_channel_parameter",
    "read_single_module_parameter",
    "read_statistics_from_module",
    "save_dsp_parameters_as_json",
    "set_DACs",
    "start_histogram_run",
    "start_list_mode_run",
    "tau_finder",
    "write_single_channel_parameter",
    "write_single_module_parameter",
    "converter_IEEE754_to_ulong",
    "converter_ulong_to_IEEE754",
]

from collections.abc import Iterable
import ctypes
import logging
from pathlib import Path
import sys
from typing import Dict, List, Union, Optional
from typing import Iterable as Iterable_type
from unittest.mock import MagicMock

import numpy as np

from . import config

# set up logging
log = logging.getLogger(__name__)

SDK = None


def load_library() -> bool:
    """Loads the C-library or a MagidMock in case the library is not available.

    Returns True on success, else False.
    """
    global SDK
    try:
        SDK = ctypes.cdll.LoadLibrary(config.get_sdk())

        for func in [
            SDK.Pixie16ComputeRealTime,
            SDK.Pixie16ComputeInputCountRate,
            SDK.Pixie16ComputeLiveTime,
            SDK.Pixie16ComputeOutputCountRate,
            SDK.Pixie16ComputeProcessedEvents,
            SDK.Pixie16ComputeRawInputCount,
            SDK.Pixie16ComputeRawOutputCount,
        ]:
            func.restype = ctypes.c_double
        return True

    except:
        SDK = MagicMock(name="Mock of XIA's C-library")
        return False


valid_module_parameter_names = [
    "MODULE_CSRA",
    "MODULE_CSRB",
    "MODULE_FORMAT",
    "MAX_EVENTS",
    "SYNCH_WAIT",
    "IN_SYNCH",
    "SLOW_FILTER_RANGE",
    "FAST_FILTER_RANGE",
    "FastTrigBackplaneEna",
    "CrateID",
    "SlotID",
    "ModID",
    "TrigConfig0",
    "TrigConfig1",
    "TrigConfig2",
    "TrigConfig3",
    "HOST_RT_PRESET",
]

valid_channel_parameter_names = [
    "TRIGGER_RISETIME",
    "TRIGGER_FLATTOP",
    "TRIGGER_THRESHOLD",
    "ENERGY_RISETIME",
    "ENERGY_FLATTOP",
    "TAU",
    "TRACE_LENGTH",
    "TRACE_DELAY",
    "VOFFSET",
    "XDT",
    "BASELINE_PERCENT",
    "EMIN",
    "BINFACTOR",
    "BASELINE_AVERAGE",
    "CHANNEL_CSRA",
    "CHANNEL_CSRB",
    "BLCUT",
    "INTEGRATOR",
    "FASTTRIGBACKLEN",
    "CFDDelay",
    "CFDScale",
    "CFDThresh",
    "QDCLen0",
    "QDCLen1",
    "QDCLen2",
    "QDCLen3",
    "QDCLen4",
    "QDCLen5",
    "QDCLen6",
    "QDCLen7",
    "ExtTrigStretch",
    "VetoStretch",
    "MultiplicityMaskL",
    "MultiplicityMaskH",
    "ExternDelayLen",
    "FtrigoutDelay",
    "ChanTrigStretch",
]


def converter_IEEE754_to_ulong(x: float) -> ctypes.c_ulong:
    """Converts a floating point number to a 32 bit integers representation."""
    a = (ctypes.c_float * 1)(x)
    b = ctypes.cast(a, ctypes.POINTER(ctypes.c_ulong))
    return b.contents


def converter_ulong_to_IEEE754(x: ctypes.c_ulong) -> float:
    """Converts a 32bit integer into a float."""
    a = (ctypes.c_ulong * 1)(x)
    b = ctypes.cast(a, ctypes.POINTER(ctypes.c_float))
    return b.contents


def check_return_code(return_code: int, function: str) -> None:
    """Checks that the API call's return code didn't return an error.

    Note
    ----
    Copied from XIA's python example.

    Parameters
    ----------
    return_code
        Value returned by XIA SDK library
    function
        Name of the function that produced the error code
    """

    if isinstance(return_code, MagicMock):
        return
    if return_code < 0:
        msg = ctypes.create_string_buffer(1024, 1024)
        SDK.PixieGetReturnCodeText(return_code, msg, 1024)

        msg = "".join(x.decode() for x in msg if x).strip()
        log.error(f"{function} failed with Error Code {return_code} : {msg}.")


def acquire_adc_trace(module: int) -> List[np.ndarray]:
    """Collects ADC traces for all 16 channels in a specific module.

    Parameters
    ----------
    module
       Module number to collect traces from

    Returns
    -------
    List[np.ndarray]
       list containing 16 traces as numpy arrays
    """
    Cmodule = ctypes.c_ushort(module)

    ret = SDK.Pixie16AcquireADCTrace(Cmodule)
    check_return_code(ret, "Pixie16AcquireADCTrace")

    traces = []
    for ch in range(16):
        Cch = ctypes.c_ushort(ch)
        trc_length = ctypes.c_uint(0)
        ret = SDK.PixieGetTraceLength(Cmodule, Cch, ctypes.byref(trc_length))
        check_return_code(ret, "PixieGetTraceLength")

        trace = (ctypes.c_ushort * trc_length.value)()
        ret = SDK.Pixie16ReadSglChanADCTrace(
            ctypes.cast(trace, ctypes.POINTER(ctypes.c_ushort * 1)),
            trc_length,
            Cmodule,
            Cch,
        )
        check_return_code(ret, "Pixie16ReadSglChanADCTrace")

        trace = np.ctypeslib.as_array(trace)

        traces.append(trace)

    return traces


def adjust_offsets(module: int) -> None:
    """Executes the adjust_offsets control task for a module.

    Parameters
    ----------
    module
       Module number to collect traces from
    """
    Cmodule = ctypes.c_ushort(module)

    ret = SDK.Pixie16AdjustOffsets(Cmodule)
    check_return_code(ret, "Pixie16AdjustOffsets")


def baseline_cut_finder(module: int, channel: int) -> None:
    """Find an appropriate baseline cuts for a channel in a module.

    Parameters
    ----------
    module
       Module number
    channel
       Channel number
    """
    Cmodule = ctypes.c_ushort(module)

    Cch = ctypes.c_ushort(channel)
    val = ctypes.c_uint32(0)
    ret = SDK.Pixie16BLcutFinder(Cmodule, Cch, ctypes.byref(val))
    check_return_code(ret, "Pixie16BLcutFinder")
    return ret


def boot_module(
    DSPParFile: Path,
    ComFPGAConfigFile: Path,
    SPFPGAConfigFile: Path,
    DSPCodeFile: Path,
    DSPVarFile: Path,
    ModNum: int,
    BootPattern: int = 0x7F,
    verbose: bool = False,
) -> None:
    """Boot the module(s).

    See page 13 of the programming manual.


    Parameters
    ----------
    ComFPGAConfigFile
         config file found under Firmware
    SPFPGAConfigFile
         config file found under Firmware
    DSPCodeFile
         config file found under DSP
    DSPParFile
         config file found under Configuration
    DSPVarFile
         config file found under DSP
    ModNum
         location of module you want to boot
         (either 0,...,k-1 for individual modules or k for all modules)
         if None, check global modules variable to boot all modules
    BootPattern
        boot pattern mask. 0x7F to boot all on-board chips.
    verbose
        print out what firmware we are using

    Returns
    -------
    bool
       return True on success and False on error
    """
    for f in [ComFPGAConfigFile, SPFPGAConfigFile, DSPCodeFile, DSPVarFile]:
        if not f.is_file():
            print(f"[Error] cannot open firmware file {f}")
            sys.exit(1)

    # config option is not used anymore, just pass an empty string
    TrigFPGAConfigFile = " "

    if verbose:
        print("Booting Pixie using the following firmware:")
        print(f"  Com      = {ComFPGAConfigFile}")
        print(f"  SP       = {SPFPGAConfigFile}")
        print(f"  DSP Code = {DSPCodeFile}")
        print(f"  DSP Var  = {DSPVarFile}")
        print(f"  Setting  = {DSPParFile}")

    # convert to ctypes for library call
    CComFPGAConfigFile = ctypes.c_char_p(bytes(ComFPGAConfigFile))
    CSPFPGAConfigFile = ctypes.c_char_p(bytes(SPFPGAConfigFile))
    # converting a string next, so we need to specify utf8
    CTrigFPGAConfigFile = ctypes.c_char_p(bytes(TrigFPGAConfigFile, "utf8"))
    CDSPCodeFile = ctypes.c_char_p(bytes(DSPCodeFile))
    CDSPParFile = ctypes.c_char_p(bytes(DSPParFile))
    CDSPVarFile = ctypes.c_char_p(bytes(DSPVarFile))
    CModNum = ctypes.c_ushort(ModNum)
    CBootPattern = ctypes.c_ushort(BootPattern)

    ret = SDK.Pixie16BootModule(
        CComFPGAConfigFile,
        CSPFPGAConfigFile,
        CTrigFPGAConfigFile,
        CDSPCodeFile,
        CDSPParFile,
        CDSPVarFile,
        CModNum,
        CBootPattern,
    )
    check_return_code(ret, "boot_module")
    if ret == 0:
        log.debug("Boot Success!")

    return


def check_external_FIFO_status(module: int) -> int:
    """Read how many 32bit words are available in the FIFO of a specific module.

    Parameters
    ----------
    module
       module number

    Returns
    -------
    int
       number of 32 bit words available
    """
    Cmodule = ctypes.c_ushort(module)
    Cwords = (ctypes.c_uint)()

    ret = SDK.Pixie16CheckExternalFIFOStatus(ctypes.byref(Cwords), Cmodule)
    check_return_code(ret, "check_external_FIFO_status")

    return Cwords.value


def check_run_status(module: int) -> bool:

    """Check the run status of a module.

    See page 17 in the manual.

    Parameters
    ----------
    module
       module number

    Returns
    -------
    bool or None
       True if run is still ongoing, False if not and None on error.
    """
    Cmodule = ctypes.c_ushort(module)

    ret = SDK.Pixie16CheckRunStatus(Cmodule)
    check_return_code(ret, "check_run_status")

    if ret == 1:
        return True
    if ret == 0:
        return False
    # ret < 0 indicates an error
    return False


def copy_DSP_parameters(
    bitmask: int,
    source_module: int,
    source_channel: int,
    destination_mask: List[List[bool]],
) -> None:
    """Copy DSP parameters from one module to other modules.

    The bit mask allows to select 13 different parameter groups that will be copied or not:

    .. list-table:: Functions
       :widths: 25 25
       :header-rows: 1

       * - Bit
         - Item
       * - 0
         - Filter (energy and trigger)
       * - 1
         - Analog signal conditioning (polarity, dc-offset, gain/attenuation)
       * - 2
         - Histogram control (minimum energy, binning factor)
       * - 3
         - Decay time
       * - 4
         - Pulse shape analysis (trace length and trace delay)
       * - 5
         - Baseline control (baseline cut, baseline percentage)
       * - 7
         - Channel CSRA register (good channel, trigger enabled, etc.)
       * - 8
         - CFD trigger (CFD delay, scaling factor)
       * - 9
         - Trigger stretch lengths (veto, external trigger, etc.)
       * - 10
         - FIFO delays (analog input delay, fast trigger output delay, etc.)
       * - 11
         - Multiplicity (bit masks, thresholds, etc.)
       * - 12
         - QDC (QDC sum lengths)

    Paramters
    ---------
    bitmask
        Bit pattern of what to copy, see above
    source_module
        The module number to copy from
    source_channel
        The source channel
    destination_mask
        List of List. For each module in the system a List of 16 True/False
        values that indicate if the settings should be copied to this module
        and this channel.
    """
    # flatten the list
    destination_mask = [ch_bit for module in destination_mask for ch_bit in module]

    CModNum = ctypes.c_ushort(source_module)
    CChannel = ctypes.c_ushort(source_channel)
    Cbitmask = ctypes.c_ushort(bitmask)
    CDestinationMask = (ctypes.c_ushort * len(destination_mask))()
    for i, value in enumerate(destination_mask):
        CDestinationMask[i] = value

    ret = SDK.Pixie16CopyDSPParameters(Cbitmask, CModNum, CChannel, CDestinationMask)
    check_return_code(ret, "copy_DSP_parameters")


def end_run(module: int) -> None:
    """End the current measurement run in a specific module.

    Parameters
    ----------
    module
       module number
    """
    CModNum = ctypes.c_ushort(module)

    ret = SDK.Pixie16EndRun(CModNum)
    check_return_code(ret, "end_run")


def exit_system(module: int) -> None:
    """Release resources used by the modules before exiting the application.

    For k modules, the number is either 0,...,k-1 for individual
    modules or k for all modules.

    Parameters
    ----------
    module
       module number
    """
    CModNum = ctypes.c_ushort(module)

    ret = SDK.Pixie16ExitSystem(CModNum)
    check_return_code(ret, "exit_system")

    if ret == 0:
        log.debug("Exit system Success!")


def init_system(
    modules: Optional[List[int]] = None,
    offline_mode: bool = False,
    verbose: bool = True,
) -> None:
    """Initialize the system.

    Parameters
    ----------
    modules
       array containing the slot numbers of each module
    offline_mode
       specify to use online or offline mode
    verbose
       if True, print more output

    Returns
    -------
    bool
       returns True if system initialized or False otherwise
    """
    if modules is None:
        modules = [2]

    if verbose:
        print(f"[INFO] Using modules in slot(s) {' '.join([str(x) for x in modules])}.")

    NumModules = len(modules)
    CNumModules = ctypes.c_ushort(NumModules)
    CPXISlotMap = (ctypes.c_ushort * (NumModules))()
    for i, slot in enumerate(modules):
        CPXISlotMap[i] = ctypes.c_ushort(slot)
    COfflineMode = ctypes.c_ushort(offline_mode)
    ret = SDK.Pixie16InitSystem(CNumModules, CPXISlotMap, COfflineMode)
    check_return_code(ret, "init_system")

    if ret == 0:
        log.debug("Initialize Success!")

    return


def load_dsp_parameters_from_file(filename: Path) -> None:
    """Read the DSP parameter from either an old .set file or a json file.

    Parameters
    ----------
    filename
       DSP parameter file (either old .set or new .json)
    """
    if not filename.exists():
        log.error(f"Setting file {filename} does not exist.")
        return
    if filename.suffix not in [".set", ".json"]:
        log.error(f"Setting file {filename} needs to be either .set or .json.")
        return

    Cfilename = ctypes.c_char_p(bytes(str(filename.absolute()), "utf8"))

    ret = SDK.Pixie16LoadDSPParametersFromFile(Cfilename)
    check_return_code(ret, "load_dsp_parameters_from_file")


def read_data_from_external_FIFO(module: int, words: int) -> np.ndarray:
    """Read data in 32bit words from module N.

    Parameters
    ----------
    module
        module number
    words
        number of words to read

    Returns
    -------
    np.ndarray
         numpy array of 32 bit words (unsigned integers 32bit)
    """
    Cmodule = ctypes.c_ushort(module)
    Cwords = ctypes.c_uint(words)
    Cdata = (ctypes.c_uint * words)()

    ret = SDK.Pixie16ReadDataFromExternalFIFO(Cdata, Cwords, Cmodule)
    check_return_code(ret, "read_data_from_external_FIFO")

    return np.ctypeslib.as_array(Cdata)


def read_histogram_from_module(module: int, channel: int, N: int = 32768) -> np.ndarray:
    """Read a single channel histogram from the module.

    Parameters
    ----------
    module
       module number
    channel
       channel in module
    N
       number of bins in the histogram, by default 32768

    Returns
    -------
    np.ndarray
       1D-numpy array with the histogram data
    """
    Cmodule = ctypes.c_ushort(module)
    Cchannel = ctypes.c_ushort(channel)
    result = (ctypes.c_uint * N)()
    N = ctypes.c_uint(N)

    ret = SDK.Pixie16ReadHistogramFromModule(result, N, Cmodule, Cchannel)
    check_return_code(ret, "read_histogram_from_module")

    return np.ctypeslib.as_array(result)


def read_module_info(module: int) -> Dict:
    """Read the module information.

    Parameters
    ----------
    module
       The module number

    Results
    -------
    dict
       dictionary with serial number, revision and other information
    """
    Cmodule = ctypes.c_ushort(module)

    rev = ctypes.c_ushort()
    sn = ctypes.c_uint()
    adc_bits = ctypes.c_ushort()
    adc_msps = ctypes.c_ushort()
    num_channels = ctypes.c_ushort()

    ret = SDK.Pixie16ReadModuleInfo(
        Cmodule,
        ctypes.byref(rev),
        ctypes.byref(sn),
        ctypes.byref(adc_bits),
        ctypes.byref(adc_msps),
        ctypes.byref(num_channels),
    )
    check_return_code(ret, "read_module_info")

    return {
        "module": module,
        "serial number": sn.value,
        "rev": rev.value,
        "adc_bits": adc_bits.value,
        "adc_msps": adc_msps.value,
        "num_channels": num_channels.value,
    }


def read_baselines(module: int) -> List[List[np.ndarray]]:
    """Acquire and read back baselines for all channels in a module.

    Parameters
    ----------
    module
       The module number

    Returns
    -------
    List[np.ndarray]
       Baselines for each channels in the module
    """
    Cmodule = ctypes.c_ushort(module)

    ret = SDK.Pixie16AcquireBaselines(Cmodule)
    check_return_code(ret, "Pixie16AcquireBaselines")

    baselines = []
    timestamps = []
    max_num_baselines = 3640

    Cmax_num_baselines = ctypes.c_ushort(max_num_baselines)

    for ch in range(16):
        Cch = ctypes.c_ushort(ch)
        bl = (ctypes.c_double * max_num_baselines)()
        ts = (ctypes.c_double * max_num_baselines)()

        ret = SDK.Pixie16ReadSglChanBaselines(
            ctypes.cast(bl, ctypes.POINTER(ctypes.c_double * 1)),
            ctypes.cast(ts, ctypes.POINTER(ctypes.c_double * 1)),
            Cmax_num_baselines,
            Cmodule,
            Cch,
        )
        check_return_code(ret, "Pixie16ReadSglChanBaselines")

        bl = np.ctypeslib.as_array(bl)
        ts = np.ctypeslib.as_array(ts)

        baselines.append(bl)
        timestamps.append(ts)
    return [baselines, timestamps]


def read_single_channel_parameter(
    channel_parameter_name: str,
    module: int,
    channel: int,
    convert_bit_pattern: bool = True,
) -> Union[List[bool], float]:
    """Read a parameter in one channel in a module.

    See pg. 51 of the programmers manual for a list of parameters available.

    Parameters
    ----------
    channel_parameter_name
        parameter name
    module
        module number
    channel
        channel number

    Returns
    -------
    float or List[bools]
        Returns the value as floating number or for bitmasks as list of bools
    """

    assert (
        channel_parameter_name in valid_channel_parameter_names
    ), "Not a valid channel parameter name"

    # only convert certain settings to an array of bits (if requested)
    if convert_bit_pattern and channel_parameter_name not in [
        "CHANNEL_CSRA",
        "CHANNEL_CSRB",
        "MultiplicityMaskL",
        "MultiplicityMaskH",
    ]:
        convert_bit_pattern = False

    # convert to ctypes for library call
    CChanParName = ctypes.c_char_p(bytes(channel_parameter_name, "utf8"))
    Cmodule = ctypes.c_ushort(module)
    Cchannel = ctypes.c_ushort(channel)
    CChanParData = (ctypes.c_double)()

    ret = SDK.Pixie16ReadSglChanPar(
        CChanParName, ctypes.byref(CChanParData), Cmodule, Cchannel
    )
    check_return_code(ret, "read_single_channel_parameter")

    if convert_bit_pattern:
        value = np.ctypeslib.as_array(CChanParData)
        value = f"{int(value):032b}"
        value = [bool(int(x)) for x in value]
        return value

    return np.ctypeslib.as_array(CChanParData)


def read_single_module_parameter(
    module_parameter_name: str, module: int, convert_bit_pattern: Optional[bool] = False
) -> Union[int, float, List[bool]]:
    """Read a parameter in one module.

    See pg. 53 of the programmers manual for a list of parameters available.

    Parameters
    ----------
    module_parameter_name
        parameter name
    module
        module number

    Returns
    -------
    int or float
        The value of the parameter
    """
    assert (
        module_parameter_name in valid_module_parameter_names
    ), "Wrong  parameter name"

    # only convert certain settings to an array of bits (if requested)
    if convert_bit_pattern and module_parameter_name not in [
        "MODULE_CSRA",
        "MODULE_CSRB",
        "FastTrigBackplaneEna",
        "TrigConfig0",
        "TrigConfig1",
        "TrigConfig2",
        "TrigConfig3",
    ]:
        convert_bit_pattern = False

    # which parameters do we need to convert to float?
    parameter_list_convert_to_IEEE = ["HOST_RT_PRESET"]

    # convert to ctypes for library call
    CModParName = ctypes.c_char_p(bytes(module_parameter_name, "utf8"))
    Cmodule = ctypes.c_ushort(module)
    CModParData = (ctypes.c_uint)()

    ret = SDK.Pixie16ReadSglModPar(CModParName, ctypes.byref(CModParData), Cmodule)
    check_return_code(ret, "read_single_module_parameter")

    if convert_bit_pattern:
        value = np.ctypeslib.as_array(CModParData)
        value = f"{int(value):032b}"
        value = [bool(int(x)) for x in value]
        return value

    if module_parameter_name in parameter_list_convert_to_IEEE:
        return np.ctypeslib.as_array(converter_ulong_to_IEEE754(CModParData))

    return np.ctypeslib.as_array(CModParData)


def read_statistics_from_module(module: int) -> Dict:
    """Read statistics from a module.

    Parameters
    ----------
    module
       module number

    Returns
    -------
    dict
       Return stats for one module
    """
    Cmodule = ctypes.c_ushort(module)

    bin_stats = (ctypes.c_uint * SDK.Pixie16GetStatisticsSize())()
    ret = SDK.Pixie16ReadStatisticsFromModule(
        ctypes.cast(bin_stats, ctypes.POINTER(ctypes.c_uint * 1)), Cmodule
    )
    check_return_code(ret, "read_statistics_from_module")

    stats_labels = (
        "real_time",
        "live_time",
        "input_counts",
        "input_count_rate",
        "output_counts",
        "output_count_rate",
    )

    stats = [[0] * 16 for x in range(len(stats_labels))]

    for ch in range(16):
        stats[0][ch] = SDK.Pixie16ComputeRealTime(bin_stats, Cmodule)
        stats[1][ch] = SDK.Pixie16ComputeLiveTime(bin_stats, Cmodule, ch)
        stats[2][ch] = SDK.Pixie16ComputeRawInputCount(bin_stats, Cmodule, ch)
        stats[3][ch] = SDK.Pixie16ComputeInputCountRate(bin_stats, Cmodule, ch)
        stats[4][ch] = SDK.Pixie16ComputeRawOutputCount(bin_stats, Cmodule, ch)
        stats[5][ch] = SDK.Pixie16ComputeOutputCountRate(bin_stats, Cmodule, ch)
        logging.info(
            {
                "module": module,
                "channel": ch,
                "real_time": stats[0][ch],
                "live_time": stats[1][ch],
                "input_counts": stats[2][ch],
                "input_count_rate": stats[3][ch],
                "output_counts": stats[4][ch],
                "output_count_rate": stats[5][ch],
            }
        )
    out = {
        "module": module,
        "real_time": stats[0],
        "live_time": stats[1],
        "input_counts": stats[2],
        "input_count_rate": stats[3],
        "output_counts": stats[4],
        "output_count_rate": stats[5],
    }

    return out


def save_dsp_parameters_as_json(filename: Path) -> None:
    """Save the current DSP parameters to a file.

    Uses the new json style.

    Parameters
    ----------
    Filename
        DSP parameter file name (with complete path)
    """
    if not filename.suffix == ".json":
        log.error("Filename does not end in .json for saving settings")
        return
    Cfilename = ctypes.c_char_p(bytes(str(filename.absolute()), "utf8"))

    ret = SDK.Pixie16SaveDSPParametersToFile(Cfilename)
    check_return_code(ret, "save_dsp_parameters_as_json")


def set_DACs(module: int) -> None:
    """Update voltage offsets from the current setting

    The new setting already needs to be set. This just activates it.

    Parameters
    ----------
    module
        the module that will be updated
    """
    Cmod = ctypes.c_ushort(module)

    ret = SDK.Pixie16SetDACs(Cmod)
    check_return_code(ret, "set_DACs")


def start_histogram_run(module: int, resume: bool = False) -> None:
    """Start an MCA run in a module.

    Parameters
    ----------
    module
         module number
         0,..., k-1 to start individual modules, k to start all modules
    resume
         Resume the run or clear old histograms
    """
    Cmodule = ctypes.c_ushort(module)
    Cmode = ctypes.c_ushort(not resume)  # 1 = start new run, 0 = resume old

    ret = SDK.Pixie16StartHistogramRun(Cmodule, Cmode)
    check_return_code(ret, "start_histogram_run")


def start_list_mode_run(module: int, resume: bool = False) -> None:
    """Start a list mode run in one channel in a module.

    Parameters
    ----------
    module
         module number
         0,..., k-1 to start individual modules, k to start all modules
    resume
         If `False` erase histogram and statistics information
    """
    Cmodule = ctypes.c_ushort(module)
    CrunType = ctypes.c_ushort(0x100)
    Cmode = ctypes.c_ushort(not resume)  # 1 = start new run, 0 = resume old

    ret = SDK.Pixie16StartListModeRun(Cmodule, CrunType, Cmode)
    check_return_code(ret, "start_list_mode_run")


def tau_finder(module: int) -> np.ndarray:
    """Find the decay constants.

    Parameters
    ----------
    module
       module number

    Returns
    -------
    np.ndarray
       The values of the decay constant for each channel, -1 if no tau has been found
    """
    Cmodule = ctypes.c_ushort(module)
    Ctaus = (ctypes.c_double * 16)()

    ret = SDK.Pixie16TauFinder(
        Cmodule,
        ctypes.cast(Ctaus, ctypes.POINTER(ctypes.c_double * 1)),
    )
    check_return_code(ret, "tau_finder")

    return np.ctypeslib.as_array(Ctaus)


def write_single_channel_parameter(
    channel_parameter_name: str,
    channel_parameter_data: Union[int, float, Iterable_type[bool]],
    module: int,
    channel: int,
) -> None:
    """Change a parameter in one channel in a module.

    See pg. 67 of the programmers manual for a list of parameters available.

    Parameters
    ----------
    channel_parameter_name
        parameter name
    channel_parameter_data
        value of the parameter you wish to set
    module
        module number
    channel
        channel number
    """
    assert (
        channel_parameter_name in valid_channel_parameter_names
    ), f"Not a valid channel parameter name. Given value {channel_parameter_name}"

    if channel_parameter_name in [
        "CHANNEL_CSRA",
        "CHANNEL_CSRB",
        "MultiplicityMaskL",
        "MultiplicityMaskH",
    ]:
        if isinstance(channel_parameter_data, (list, tuple)):
            data = [str(int(x)) for x in channel_parameter_data]
            channel_parameter_data = int("".join(data), 2)

    # We encountered a bug when setting coincidence windows to 10ns, so we check here and print a warning
    if channel_parameter_name == "FASTTRIGBACKLEN":
        if channel_parameter_data == 0.01:
            print(
                f"[red]CRITICAL[/] :boom: :boom: :boom:The coincidence window for mod={module}:ch={channel} was set to 10 ns. "
                f"This could lead to data loss. :boom: :boom: :boom:"
            )

    # convert to ctypes for library call
    CChanParName = ctypes.c_char_p(bytes(channel_parameter_name, "utf8"))
    CChanParData = ctypes.c_double(channel_parameter_data)
    Cmodule = ctypes.c_ushort(module)
    Cchannel = ctypes.c_ushort(channel)

    ret = SDK.Pixie16WriteSglChanPar(CChanParName, CChanParData, Cmodule, Cchannel)
    check_return_code(ret, "write_single_channel_parameter")
    if ret < 0:
        print(
            f"   Problem in mod={module} ch={channel} name={channel_parameter_name} value={channel_parameter_data}"
        )


def write_single_module_parameter(
    module_parameter_name: str,
    module_parameter_data: Union[int, float, List[bool]],
    module: int,
) -> None:
    """Change a parameter in one module.

    See pg. 69 of the programmers manual for a list of parameters available.

    Parameters
    ----------
    module_parameter_name
        parameter name
    module_parameter_data
        value of the parameter you wish to set
    module
        module number
    """

    assert (
        module_parameter_name in valid_module_parameter_names
    ), f"Module parameter name is invalid. Given module parameter name: {module_parameter_name}"
    assert isinstance(module_parameter_data, (int, float)) or all(
        isinstance(x, bool) for x in module_parameter_data
    ), f"Module parameter '{module_parameter_name}' not an integer or float. Given value is {module_parameter_data}"
    # which parameters do we need to convert from float?
    parameter_list_convert_from_IEEE = ["HOST_RT_PRESET"]

    if module_parameter_name in [
        "MODULE_CSRA",
        "MODULE_CSRB",
    ]:
        if isinstance(module_parameter_data, Iterable):
            data = [str(int(x)) for x in module_parameter_data]
            module_parameter_data = int("".join(data), 2)

    # convert to ctypes for library call
    CModParName = ctypes.c_char_p(bytes(module_parameter_name, "utf8"))
    if module_parameter_name in parameter_list_convert_from_IEEE:
        CModParData = converter_IEEE754_to_ulong(module_parameter_data)
    else:
        CModParData = ctypes.c_uint(module_parameter_data)
    Cmodule = ctypes.c_ushort(module)

    ret = SDK.Pixie16WriteSglModPar(CModParName, CModParData, Cmodule)
    check_return_code(ret, "write_single_module_parameter")
    if ret != 0:
        print(
            f" write_single_module_parameter error for {module_parameter_name=} {module_parameter_data=} {module=}"
        )



"""
Functions to help analyze PIXIE-16 data
"""

from functools import wraps

import numpy as np


def ensure_2d_array(f):
    """Convert 1D numpy array to 2D arrays and back.

    Using this decorator ensures that the first argument is always a 2d array.
    This way all functions below only need to handle that case.

    """

    @wraps(f)
    def wrap(*args, **kwargs):
        # check if first argument is accessible and a numpy array
        try:
            array = args[0]
        except IndexError:
            raise TypeError(
                f"[ERROR] {f.__name__}: first argument needs to be a numpy array"
            )
        if not isinstance(array, np.ndarray):
            raise TypeError(
                f"[ERROR] {f.__name__}: first argument needs to be a numpy array"
            )

        # convert 1D to 2D if needed
        if len(array.shape) == 1:
            array = array[np.newaxis, :]
            args = list(args)
            args[0] = array
            one_dim = True
        else:
            one_dim = False

        # call orginal function now with a 2D array
        out = f(*args, **kwargs)

        # convert 2D array back to 1D if needed
        if one_dim:
            if isinstance(out, np.ndarray):
                out = out[0]
            elif isinstance(out, (list, tuple)):
                out = (o[0] for o in out)
        return out

    return wrap


@ensure_2d_array
def convert_trace_to_FPGA(traces: np.ndarray, filter_range: int = 0):
    """Sum up 2**filter_range steps in trace to match FPGA clock cycle.

    The FPGA (100 MHz = 10 ns) always handels at least 5 data points
    (2ns) at a time. It does this by adding up 5 data points into a
    single point.

    For some calculations the FPGA sums up more data points (up to
    2^filter_range), for example, for the energy calculations the minimum
    filter_range setting is 1.

    Parameters
    ----------
    traces
        A 1D or 2D array of 2 ns time step data. For 1D, the number of
        data points, M, needs to be a multiple of 5*2^filter_range. For 2D, the
        size N,M of the array needs to confirm to N traces with a
        length of M, where M needs to fulfill the same constraints as
        in the 1D case.
    filter_range
        5*2^filter_range points will be added up

    Result
    ------
        For the 1D case, return a ndarray with size M/(5*2^filter_range) and
        for the 2D case, return a N,M/(5*2^filter_range) array

    """
    N, M = traces.shape

    if filter_range < 0:
        raise ValueError("The filter_range parameter needs to be >= 0")

    group = 5 * 2 ** filter_range

    if M % group != 0:
        raise ValueError(
            f"The trace length is not a multiple of 5*2^filter_range, for filter_range={filter_range}"
        )

    return traces.reshape(N, M // group, group).sum(axis=2)


@ensure_2d_array
def calculate_CFD(traces, t=None, CFD_threshold=None, w=1, B=5, D=5, L=1, Nbkgd=100):
    """Calculate the CFD trace as is done by the PIXIE

    Also provide the option to change parameters used in the calculations
    (The calculations we do here is for the 500 MHz version).

    The input can be 1d or 2d in which case the output will be also 1d or 2d.

    input:
       traces   (N, M) numpy array: N= number of traces, M=number of data points
       t        (M, 1) numpy arrray, the time axes for the traces

    output:
       CFD  (N, M-B-D-L+1)  numpy array
       zero_crossing        (N, 1) numpy array: the time of the zero crossing
       errors               True/False array if CFD could be established or not

    """
    traces = traces.astype(np.int64)
    NR_traces, length = traces.shape
    errors = np.zeros(NR_traces, dtype=np.bool)
    cfdtrigger = np.zeros(NR_traces)

    # we are skipping the first B+D data points, so the CFD array will
    # be shorter, adjusting the time array accordingly
    if t is None:
        t2 = np.arange(0, length - B - D) * 2e-9
    else:
        t2 = t[B + D :]

    # create an empty array of the correct length
    CFD = np.zeros((NR_traces, length - B - D - L + 1))
    cfdtime = []
    # calculate the CFD, see section 3.3.8.2, page 47 of the PIXIE
    # manual do as much as possible using numpy, one probably can and
    # should get rid of the for-loop here too, just not sure at the
    # moment how to do it
    for k in range(B + D, length - L):
        CFD[:, k - B - D] = w * (
            traces[:, k : k + L + 1].sum(axis=1)
            - traces[:, k - B : k - B + L + 1].sum(axis=1)
        ) - (
            traces[:, k - D : k - D + L + 1].sum(axis=1)
            - traces[:, k - D - B : k - D - B + L + 1].sum(axis=1)
        )

    # now find the zero crossing
    for i, CFDtrace in enumerate(CFD):
        # check that the first N points (noise) is below threshold
        # 100 is an arbitrary number here, but the PIXIE should always record some noise first
        if CFDtrace[:Nbkgd].max() > CFD_threshold:
            print("Warning: CFD analysis theshold too low?", CFDtrace[:Nbkgd].max())

        try:
            # all zero crossings
            zero_crossings = np.where(np.diff(np.signbit(CFDtrace)))[0]

            # find first index where the CFD signal is above the threshold
            CFD_trigger_index = np.where(CFDtrace >= CFD_threshold)[0][0]

            # first zero crossing at later time than the threshold
            left = zero_crossings[zero_crossings > CFD_trigger_index][0]
            right = left + 1

            # sanity check
            if CFDtrace[left] > CFDtrace[right]:
                zer = t2[left] + CFDtrace[left] * (t2[right] - t2[left]) / (
                    CFDtrace[left] - CFDtrace[right]
                )
            else:
                print(
                    "CFD Error: not at a zero crossing from positive to negative values "
                )
                print("   This should not happen!!")
                print("   Trace index:", i)
                print("   CFD values:", CFDtrace[left], CFDtrace[right])
                zer = t2[left]
        except IndexError:
            errors[i] = True
        else:
            cfdtrigger[i] = zer

    return CFD, cfdtime, errors


@ensure_2d_array
def calculate_CFD_using_FF(
    traces,
    t=None,
    CFD_threshold=None,
    FF_threshold=None,
    Lf=10,
    Gf=10,
    w=1,
    B=5,
    D=5,
    L=1,
    Nbkgd=10,
    FF_delay=0,
    CFD_delay=0,
):
    """Calculate the CFD trace as is done by the PIXIE

    Also provide the option to change parameters used in the calculations
    (The calculations we do here is for the 500 MHz version).

    The input can be 1d or 2d in which case the output will be also 1d or 2d.

    input:
       traces   (N, M) numpy array: N= number of traces, M=number of data points
       t        (M, 1) numpy arrray, the time axes for the traces

    output:
       CFD  (N, M-B-D-L+1)  numpy array
       cfdtime        (N, 1) numpy array: the time of the zero crossing
       FF       numpy array: fast filter used for the calculation
       IDXerr   numpy array: index of CFD errors

    """

    traces = traces.astype(np.int64)
    NR_traces, length = traces.shape

    # we are skipping the first B+D data points, so the CFD array will
    # be shorter, adjusting the time array accordingly
    if t is None:
        t2 = np.linspace(0, 2 * (traces.shape[1] - 1), traces.shape[1])  # in ns
        t2 = t2[B + D :]
    else:
        t2 = t[B + D :]

    # create an empty array of the correct length
    CFD = np.zeros((NR_traces, length - B - D - L + 1))
    cfdtime = []
    # calculate the CFD, see section 3.3.8.2, page 47 of the PIXIE
    # manual do as much as possible using numpy, one probably can and
    # should get rid of the for-loop here too, just not sure at the
    # moment how to do it
    for k in range(B + D, length - L):
        CFD[:, k - B - D] = w * (
            traces[:, k : k + L + 1].sum(axis=1)
            - traces[:, k - B : k - B + L + 1].sum(axis=1)
        ) - (
            traces[:, k - D : k - D + L + 1].sum(axis=1)
            - traces[:, k - D - B : k - D - B + L + 1].sum(axis=1)
        )

    # now find the zero crossing
    low = 0
    CFDerror = 0
    IDXerr = []
    cfdtimeIDX = []
    fast_filter = calculate_filter(traces, Lf, Gf)
    for i, (CFDtrace, FFtrace) in enumerate(zip(CFD, fast_filter)):
        FFtrace = FFtrace[FF_delay:]
        CFDtrace = CFDtrace[CFD_delay:]

        # check that the first N points of the fast filter (noise) is below threshold
        # 30 is an arbitrary number here, but the PIXIE should always record some noise first
        if (
            FFtrace[:Nbkgd].max() > FF_threshold
            or CFDtrace[:Nbkgd].max() > CFD_threshold
        ):
            # print('Warning: CFD analysis theshold too low?', CFDtrace[:Nbkgd].max(),i)
            low += 1

        # all zero crossings from poitive to negative
        zero_crossings = np.where(np.diff(np.signbit(CFDtrace).astype(np.int)) > 0)[0]
        if zero_crossings.size == 0:
            CFDerror += 1
            cfdtime.append(None)
            IDXerr.append(i)
            continue

        # find first index where the CFD trace is above the threshold
        try:
            FF_trigger_index = np.where(FFtrace >= FF_threshold)[0][0]
            CFD_trigger_index = (
                np.where(np.diff(CFDtrace >= CFD_threshold).astype(np.int) > 0)[0] + 1
            )
            CFD_trigger_index = CFD_trigger_index[CFD_trigger_index > FF_trigger_index][
                0
            ]
            left = zero_crossings[zero_crossings > CFD_trigger_index][0]
            right = left + 1
            zer = t2[left] + CFDtrace[left] * (t2[right] - t2[left]) / (
                CFDtrace[left] - CFDtrace[right]
            )
        except IndexError:
            CFDerror += 1
            cfdtime.append(None)
            IDXerr.append(i)
            continue

        cfdtime.append(zer)
        cfdtimeIDX.append(i)
    cfdtime = np.array(cfdtime)

    print("Calculated CFD errors = ", CFDerror)
    print("Events with noise above fast filter threshold = ", low)

    return CFD, cfdtime, fast_filter, IDXerr


@ensure_2d_array
def calculate_fast_filter(
    traces: np.ndarray, L: int = 1, G: int = 1, filter_range: int = 0
) -> np.ndarray:
    """Calculate the fast filter response.

    Parameters
    ----------
    traces
       1D or 2D array of 1 or N traces of lengths M.
    L
       in FPGA units, 1 = 10 ns
    G
       in FPGA units, 1 = 10 ns
    filter_range
       number of FPGA steps in 2^filter_range that are summed up

    Returns
    -------
    np.ndarray
       The calculated fast filter response of the whole trace.

    """
    assert isinstance(L, int)
    assert L > 0
    assert isinstance(G, int)
    assert G > 0

    traces = traces.astype(np.int64)

    traces = convert_trace_to_FPGA(traces, filter_range=filter_range)
    NR_traces, length = traces.shape

    # create an empty array of the correct length
    result = np.zeros((NR_traces, length - 2 * L - G + 1))

    # do as much as possible using numpy, one probably can and should
    # get rid of the for-loop here too, just not sure at the moment
    # how to do it
    for k in range(2 * L + G - 1, length):
        result[:, k - 2 * L - G + 1] = traces[:, k - L + 1 : k + 1].sum(
            axis=1
        ) - traces[:, k - 2 * L - G + 1 : k - L - G + 1].sum(axis=1)

    return result


@ensure_2d_array
def get_fast_trigger_index(traces, threshold):
    """Return the indices where the trace value is >= threshold."""

    trigger_pos = [
        np.argwhere((trace[:-1] < threshold) & (trace[1:] > threshold)).flatten() + 1
        for trace in traces
    ]
    return trigger_pos


@ensure_2d_array
def calculate_filter(traces, L=10, G=10):
    """Calculate the fast and slow filter response from the PIXIE

    This implements the filter response as discribted on page 81 in
    section 6.1 of the manual.

    The input can be 1d or 2d in which case the output will be also 1d or 2d.

    input:
       traces   (N, M) numpy array. N traces with a length of M
       L        integer: length of the integration time
       G        integer: gap between the two integration region

    output:
       result   (N, M-2L-G+1)
    """
    assert (L % 5 == 0) and (
        G % 5 == 0
    ), "Filter settings needs to be multiple of 5 (FPGA clock cycles)"

    traces = traces.astype(np.int64)
    NR_traces, length = traces.shape

    assert length % 5 == 0, "Filter trace length needs to be a multiple of 5"

    # scale to 10 ns FPGA clock cycles, e.g. 5 time steps per cycle
    L = L // 5
    G = G // 5
    traces = convert_trace_to_FPGA(traces, filter_range=0)
    NR_traces, length = traces.shape

    # create an empty array of the correct length
    result = np.zeros((NR_traces, length - 2 * L - G + 1))

    # do as much as possible using numpy, one probably can and should
    # get rid of the for-loop here too, just not sure at the moment
    # how to do it
    for k in range(2 * L + G - 1, length):
        result[:, k - 2 * L - G + 1] = traces[:, k - L + 1 : k + 1].sum(
            axis=1
        ) - traces[:, k - 2 * L - G + 1 : k - L - G + 1].sum(axis=1)

    # scale back to 2 ns timesteps
    result = np.repeat(result, 5, axis=1)
    result = result / L

    return result



from collections import Counter, defaultdict
from pathlib import Path

from fast_histogram import histogram1d
import numpy as np
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QAction,
    QFileDialog,
    QTabWidget,
    QTableView,
    QVBoxLayout,
    QWidget,
)

import pixie16

from ..widgets import BaseApp
from ..helper import busy_cursor
from ..plot import Plot
from .channel_select import ChannelSelect
from .fast_trigger import FastTrigger
from .data_model import BinaryDataModel


class BinaryApp(BaseApp):
    """The main window"""

    def __init__(self, app=None):
        super().__init__(name="Pixie16 Binary File Browser", app=app)

        self.filename = None
        self.channels = []
        self.selected_channels = []
        self.data = []
        self.model = None

        FileButton = QAction(QIcon.fromTheme("document-open"), "Open", self)
        FileButton.setShortcut("Ctrl+O")
        FileButton.setStatusTip("Open binary file")
        FileButton.triggered.connect(lambda x: self.open_file())
        self.file_menu.insertAction(self.file_menu.actions()[0], FileButton)

        self.main_widget = QWidget(self)
        self.main_window = QVBoxLayout()

        self.tabs = QTabWidget()

        self.main_window.addWidget(self.tabs)
        self.main_window.addWidget(self.info_bar)

        self.main_widget.setLayout(self.main_window)
        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)

        self.datatab = QWidget()
        self.datatab_layout = QVBoxLayout()

        self.channel_select = ChannelSelect(main=self)

        self.dataview = QTableView()
        self.dataview.setSortingEnabled(True)
        self.dataview.setSelectionBehavior(QTableView.SelectRows)

        self.datatab_layout.addLayout(self.channel_select)
        self.datatab_layout.addWidget(self.dataview)
        self.datatab.setLayout(self.datatab_layout)

        self.traces = Plot(main=self)
        self.mca = Plot(main=self)
        self.fasttrigger = FastTrigger(main=self)

        self.tabs.addTab(self.datatab, "Binary data")
        self.tabs.addTab(self.traces, "Traces")
        self.tabs.addTab(self.mca, "MCA")
        self.tabs.addTab(self.fasttrigger, "Fast Trigger/CFD")

        self.tabs.currentChanged.connect(self.tab_change)

        self.show()

    def open_file(self, filename=None) -> None:
        if filename is None or not isinstance(filename, (str, Path)):
            self.filename, _ = QFileDialog.getOpenFileName(
                self, "Open file", str(self.data_path), "Binary files (*.bin)"
            )
        else:
            self.filename = filename

        # check if cancel was hit in QFileDialog
        if self.filename == "":
            self.filename = None
            return

        self.filename = Path(self.filename)
        with busy_cursor(self.app):
            self.data = list(pixie16.read.read_list_mode_data_as_events(self.filename))
            self.parse_data()
            self.set_data_model()
            self.update_gui()

        # count events per channel
        tmp = Counter([x.channel for x in self.data])
        summary = "  ".join(f"{k}:{v}" for k, v in sorted(tmp.items()))
        gamma_sum = 0
        for ch, cnt in sorted(tmp.items()):
            if ch < 9:
                gamma_sum += cnt
        gamma_sum_str = "".join(f"Gamma events: {gamma_sum}")

        if self.filename:
            self.info_bar.setText(
                f"{self.filename.name}    {summary}    {gamma_sum_str}"
            )

    def parse_data(self):
        self.channels = sorted({x.channel for x in self.data})
        self.data = sorted(self.data, key=lambda x: x.timestamp)

    def set_data_model(self):
        self.model = BinaryDataModel(self.data)
        self.dataview.setModel(self.model)

    def update_gui(self):
        self.fasttrigger.add_channels()

        self.channel_select.clear_channels()
        self.channel_select.add_channels()

    def tab_change(self, i):
        if not self.data:
            return
        if not self.model:
            return

        indexes = self.dataview.selectionModel().selectedRows()
        if i == 1:  # plot traces
            self.traces.fig.clear()
            self.traces.canvas.draw()
            # get list of channels in selections
            selected_channels = [
                self.data[index.row()].channel
                for index in sorted(indexes)
                if not self.dataview.isRowHidden(index.row())
            ]
            # make unique
            selected_channels = list(set(selected_channels))

            if not selected_channels:
                return

            self.traces.fig.clear()
            self.traces.axes = self.traces.fig.subplots(
                nrows=len(selected_channels), sharex=True
            )
            if not isinstance(self.traces.axes, np.ndarray):
                self.traces.axes = np.array([self.traces.axes])
            for index in sorted(indexes):
                row = index.row()
                if self.dataview.isRowHidden(row):
                    continue
                channel = self.data[row].channel
                trigger_pos = selected_channels.index(channel)
                self.traces.axes[trigger_pos].plot(self.data[row].trace)
            for i, channel in enumerate(selected_channels):
                self.traces.axes[i].set_title(f"channel {channel}")
            self.traces.fig.tight_layout()
            self.traces.canvas.draw()
        if i == 2:  # MCA spectra
            if self.mca.update:
                self.mca.fig.clear()
                self.mca.axes = self.mca.fig.subplots(
                    nrows=len(self.channels), sharex=True
                )
                # create list of energies for each channel
                energies = defaultdict(list)
                bins = 512
                rebin_factor = 32 * 1024 / 512
                for d in self.data:
                    energies[d.channel].append(d.energy / rebin_factor)
                X = np.linspace(0, bins - 1, bins)
                for i, [c, ax] in enumerate(zip(self.channels, self.mca.axes)):
                    tmp = histogram1d(energies[c], bins=bins, range=[0, bins - 1])
                    ax.plot(X, tmp)
                    ax.set_ylabel("Counts/bin")
                    ax.set_title(f"Channel {c}")
                ax.set_xlabel("Energy [a.u.]")
                self.mca.fig.tight_layout()
                self.mca.canvas.draw()
                self.mca.update = False

    def click_channel(self, channel, button):
        if channel == "all" and button.isChecked():
            self.selected_channels = self.channels
        elif channel == "all" and not button.isChecked():
            self.selected_channels = []
        else:
            if button.isChecked() and channel not in self.selected_channels:
                self.selected_channels.append(channel)
            elif not button.isChecked() and channel in self.selected_channels:
                self.selected_channels.remove(channel)

        self.channel_select.update_button_status()
        self.update_channel_selection()

    def update_channel_selection(self):
        for i, d in enumerate(self.data):
            if d.channel in self.selected_channels:
                self.dataview.showRow(i)
            else:
                self.dataview.hideRow(i)



import numpy as np
import pytest

import pixie16


def calculate_groups(N, groupsize):
    result = []
    j = 0
    n = 0
    for i in range(N):
        n += i
        j += 1
        if j == groupsize:
            result.append(n)
            j = 0
            n = 0
            continue
    return np.array(result)


class Ensure2DArray:
    def no_arg(self):
        @pixie16.analyze.ensures_2d_array
        def f():
            return True

        with pytest.raises(TypeError) as execinfo:
            f()

    def wrong_arg(self):
        @pixie16.analyze.ensures_2d_array
        def f(x):
            return True

        with pytest.raises(TypeError) as execinfo:
            f(0)

    def convert(self):
        @pixie16.analyze.ensures_2d_array
        def f(x):
            return len(x.shape) == 2

        assert f(np.arange(4))

    def no_convert(self):
        @pixie16.analyze.ensures_2d_array
        def f(x):
            return len(x.shape) == 2

        assert f(np.arange(10).reshape(5, 2))


class TestConvertTraceToFPGA:
    def test_group_of_1D(self):
        N = 100
        input = np.arange(N)
        for filter_range, groupsize in enumerate([5, 10, 20]):
            out = pixie16.analyze.convert_trace_to_FPGA(
                input, filter_range=filter_range
            )
            result = calculate_groups(N, groupsize)
            assert np.array_equal(out, result)

    def test_group_of_2D(self):
        M = 100
        N = 20
        input_1D = np.arange(M)
        input = np.array([input_1D for i in range(N)])
        for filter_range, groupsize in enumerate([5, 10, 20]):
            out = pixie16.analyze.convert_trace_to_FPGA(
                input, filter_range=filter_range
            )
            result_1D = calculate_groups(M, groupsize)
            result = np.array([result_1D for i in range(N)])
            assert np.array_equal(out, result)

    def test_wrong_length(self):
        with pytest.raises(ValueError) as execinfo:
            N = 99
            input = np.arange(N)
            out = pixie16.analyze.convert_trace_to_FPGA(input)
        with pytest.raises(ValueError) as execinfo:
            M = 99
            N = 20
            input_1D = np.arange(M)
            input = np.array([input_1D for i in range(N)])
            out = pixie16.analyze.convert_trace_to_FPGA(input)
        with pytest.raises(ValueError) as execinfo:
            N = 90
            input = np.arange(N)
            out = pixie16.analyze.convert_trace_to_FPGA(input, filter_range=2)

    def test_wrong_filter_range(self):
        with pytest.raises(ValueError) as execinfo:
            N = 100
            input = np.arange(N)
            out = pixie16.analyze.convert_trace_to_FPGA(input, filter_range=-1)



import sys

from . import plot
from . import ui
from . import analyze
from . import read

from ._version import version as __version__

from . import control
from . import scan
from . import pipeline
from . import tasks

from . import cli

__all__ = ["analyze", "plot", "ui", "scan", "control", "pipeline", "tasks", "cli"]



"""implementation of queues for parsing list mode data from the pixie16"""


from collections import namedtuple, deque
from itertools import islice
from pathlib import Path
from typing import Union

import cbitstruct as bitstruct
import numpy as np
import pandas as pd
from rich import print

WORD_SIZE = 4

# chunk_timestamp is a field that doesn't come from the pixie
# it can be used to for example store a unix time stamp when the data was converted
# tasks/SortEvents uses it for example this way.
Event = namedtuple(
    "Event",
    [
        "channel",
        "crate",
        "slot",
        "timestamp",
        "CFD_fraction",
        "energy",
        "trace",
        "CFD_error",
        "pileup",
        "trace_flag",
        "Esum_trailing",
        "Esum_leading",
        "Esum_gap",
        "baseline",
        "QDCSum0",
        "QDCSum1",
        "QDCSum2",
        "QDCSum3",
        "QDCSum4",
        "QDCSum5",
        "QDCSum6",
        "QDCSum7",
        "ext_timestamp",
        "chunk_timestamp",  # not a pixie16 field
    ],
    # note that defaults are aligned to the end and value without defaults are mandatory
    defaults=(-1, -1, -1, -1.0, -1, -1, -1, -1, -1, -1, -1, -1, -1.0, -1.0),
)


def byteswap_data(bytes_data):
    """returns bytes correctly byteswapped and in word sizes"""
    n_words = len(bytes_data) // WORD_SIZE
    data = bitstruct.byteswap(str(WORD_SIZE) * n_words, bytes_data)
    return data


class EmptyError(Exception):
    """To be raised if an empty queue is popped or peeked"""


class LeftoverBytesError(Exception):
    """To be raised if a queue with a partial word/event left is popped or peeked"""


class WordsIO:
    """A binary data FIFO queue/transformer.

    Put in binary data as any sized bytes or bytearray objects and
    take them out as Pixie16 word sized/formatted bytes (refer to user
    manual List Mode Data Structures section).

    """

    def __init__(self, binary_data: Union[bytes, bytearray] = b""):
        """Initialize queue"""
        self.binary_data = bytearray(binary_data)
        self.last_data = deque(maxlen=5)  # used for debugging

    def __len__(self):
        """Number of bytes in queue"""
        return len(self.binary_data)

    def is_empty(self):
        """Utility function to indicate if there is no bytes left in the queue"""
        return len(self) == 0

    def put(self, additional_binary_data: Union[bytes, bytearray]):
        """Put more bytes into queue"""
        self.binary_data.extend(additional_binary_data)

    def pop(self, size: int, raw=False):
        """Pop the top size words from queue"""
        res = self.peek(size, raw=raw)
        del self.binary_data[: size * WORD_SIZE]
        self.last_data.append(res)
        return res

    def assert_data_is_available(self, size: int):
        """Make sure `size` words are available."""
        if size * WORD_SIZE > len(self):
            raise LeftoverBytesError(
                f"Requested {size} words, only {len(self)//WORD_SIZE} words in object"
            )

    def peek(self, size: int, raw=False):
        """Get the top size words from queue without removing them.

        Return either tuple of WORD-sized, byte-swapped data (raw==False)
        or just raw bytes (raw==True).
        """
        if self.is_empty():
            raise EmptyError()

        self.assert_data_is_available(size)

        if raw:
            return self.binary_data[: size * WORD_SIZE]
        return byteswap_data(self.binary_data[: size * WORD_SIZE])


# Events in list mode consist of one mandatory group ("header") and 3 optional groups
# for energy sums, Qsums, and an external timestamp
# we define the corresponding bit patterns to be used with bitstruct here, see
# page 72 of the pixie16 manual, version 3.06
EVENTS_PARTS = {
    "first row": (
        "b1u14u5u4u4u4",
        ("pileup", "Event Length", "Header Length", "crate", "slot", "channel"),
    ),
    "header": (
        "b1u14u5u4u4u4u32u3u13u16b1u15u16",
        (
            "pileup",
            "Event Length",
            "Header Length",
            "crate",
            "slot",
            "channel",
            "EVTTIME_LO",
            "CFD trigger source bits",
            "CFD Fractional Time",
            "EVTTIME_HI",
            "trace_flag",
            "Trace Length",
            "energy",
        ),
    ),
    "energy": (
        "u32" * 3 + "f32",
        (
            "Esum_trailing",
            "Esum_leading",
            "Esum_gap",
            "baseline",
        ),
    ),
    "Qsums": (
        "u32" * 8,
        (
            "QDCSum0",
            "QDCSum1",
            "QDCSum2",
            "QDCSum3",
            "QDCSum4",
            "QDCSum5",
            "QDCSum6",
            "QDCSum7",
        ),
    ),
    "Ext Time": (
        "u32p16u16",
        (
            "Ext_TS_Lo",
            "Ext_TS_Hi",
        ),
    ),
}

# We now build all possible combinations out of these parts that can
# show up in list mode data We also add a "0" option to just read the
# first row, since this row contains the header length we need to read
HEADER_FORMATS = {
    0: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["first row"][0], EVENTS_PARTS["first row"][1]
    ),
    4: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0], EVENTS_PARTS["header"][1]
    ),
    6: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0] + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["header"][1] + EVENTS_PARTS["Ext Time"][1],
    ),
    8: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0] + EVENTS_PARTS["energy"][0],
        EVENTS_PARTS["header"][1] + EVENTS_PARTS["energy"][1],
    ),
    12: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0] + EVENTS_PARTS["Qsums"][0],
        EVENTS_PARTS["header"][1] + EVENTS_PARTS["Qsums"][1],
    ),
    16: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0]
        + EVENTS_PARTS["energy"][0]
        + EVENTS_PARTS["Qsums"][0],
        EVENTS_PARTS["header"][1]
        + EVENTS_PARTS["energy"][1]
        + EVENTS_PARTS["Qsums"][1],
    ),
    10: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0]
        + EVENTS_PARTS["energy"][0]
        + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["header"][1]
        + EVENTS_PARTS["energy"][1]
        + EVENTS_PARTS["Ext Time"][1],
    ),
    14: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0]
        + EVENTS_PARTS["Qsums"][0]
        + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["header"][1]
        + EVENTS_PARTS["Qsums"][1]
        + EVENTS_PARTS["Ext Time"][1],
    ),
    18: bitstruct.CompiledFormatDict(
        EVENTS_PARTS["header"][0]
        + EVENTS_PARTS["energy"][0]
        + EVENTS_PARTS["Qsums"][0]
        + EVENTS_PARTS["Ext Time"][0],
        EVENTS_PARTS["header"][1]
        + EVENTS_PARTS["energy"][1]
        + EVENTS_PARTS["Qsums"][1]
        + EVENTS_PARTS["Ext Time"][1],
    ),
}


def calc_cfd_error(cfd_trigger_source_bits):
    return cfd_trigger_source_bits == 7


def calc_cfd_fraction(cfd_trigger_source_bits, cfd_fractional_time):
    if calc_cfd_error(cfd_trigger_source_bits):
        return 0
    else:
        return ((cfd_trigger_source_bits - 1) + cfd_fractional_time / 8192) * 2


class ListModeDataReader:
    """A binary data FIFO queue/transformer.

    Put in binary data as any sized bytes or bytearray objects and
    take them out as Pixie16 events (refer to user manual List Mode
    Data Structures section).

    """

    def __init__(self, binary_data: Union[bytes, bytearray] = b""):
        """Initialize queue"""
        self.words_stream = WordsIO(binary_data)

    def __len__(self):
        """Number of bytes in queue"""
        return len(self.words_stream)

    def is_empty(self):
        """Utility function to indicate if there is no bytes left in the queue"""
        return self.words_stream.is_empty()

    def put(self, additional_binary_data: Union[bytes, bytearray]):
        """Put more bytes into queue"""
        self.words_stream.put(additional_binary_data)

    def pop(self) -> Event:
        """Pop the top event from queue"""

        # try to get all the words for the next event
        # if the complete next event isn't in queue an exception is thrown but no data is discarded
        try:
            event_header = HEADER_FORMATS[0].unpack(self.words_stream.peek(1))
            header_length = event_header["Header Length"]
            event_length = event_header["Event Length"]

            # ensure we have enough data, this is needed, since we
            # have two separate pop-calls (header, trace) and we need
            # to make sure that both work. Otherwise we might get
            # enough data for a header but not for the trace, in which
            # case we would add more data an try again to read a
            # header although trace data is next in the input queue
            self.words_stream.assert_data_is_available(event_length)

            # it's now save to read both header and trace
            header_words = self.words_stream.pop(header_length)
            trace_length = event_length - header_length
            if trace_length:
                trace_data = self.words_stream.pop(trace_length, raw=True)
            else:
                trace_data = None
        except LeftoverBytesError:
            raise LeftoverBytesError("Partial event left at the end of queue")

        # unpack words/trace into their field values
        try:
            event_values = HEADER_FORMATS[header_length].unpack(header_words)
        except KeyError:
            print("Something went wrong with the binary data.")
            print("Debug information:")
            print(f"   Current event header: {event_length}")
            print(
                f"   Number of bytes remaining in IO buffer:  {len(self.words_stream)}"
            )
            if len(self.words_stream):
                filename = Path(".") / "pixie_binary_rest_debug.bin"
                with filename.open("wb") as f:
                    f.write(self.words_stream.binary_data)
                print(f"   Saved IO buffer to:  {filename.absolute()}")
            for i, data in enumerate(self.words_stream.last_data):
                filename = Path(".") / f"pixie_binary_last_{i}_debug.bin"
                with filename.open("wb") as f:
                    f.write(data)
                print(f"   Saved IO last elements to:  {filename.absolute()}")
            raise

        if trace_data:
            trace = np.frombuffer(trace_data, f"({trace_length},2)<u2", count=1)[
                0, :, :
            ]
            trace = trace.flatten()
        else:
            trace = []

        # calculate event values that need to be derived from field values
        CFD_bits = event_values.pop("CFD trigger source bits")
        CFD_error = calc_cfd_error(CFD_bits)
        CFD_fraction = calc_cfd_fraction(
            CFD_bits, event_values.pop("CFD Fractional Time")
        )
        timestamp = (
            event_values.pop("EVTTIME_LO") + event_values.pop("EVTTIME_HI") * 2 ** 32
        ) * 10

        if "Ext_TS_Lo" in event_values:
            ext_timestamp = (
                event_values.pop("Ext_TS_Lo") + event_values.pop("Ext_TS_Hi") * 2 ** 32
            )

            event_values.update({"ext_timestamp": ext_timestamp})

        for key in ["Event Length", "Header Length", "Trace Length"]:
            if key in event_values:
                del event_values[key]

        event_values.update(
            {
                "timestamp": timestamp,
                "CFD_fraction": CFD_fraction,
                "trace": trace,
                "CFD_error": CFD_error,
            }
        )

        return Event(**event_values)

    def iterevents(self):
        while True:
            try:
                yield self.pop()
            except (EmptyError, LeftoverBytesError):
                return
            except TypeError as e:
                print(f"TypeError: {e}", flush=True)
                return

    def pop_all(self):
        return tuple(self.iterevents())


class FilesIO:
    """An object to stream bytes from a list of files"""

    def __init__(self, files):
        self.files = files
        self.current_file_index = 0
        self.current_file = None
        self._is_empty = False

    def is_empty(self):
        """Utility function to indicate if there is no bytes left in the queue"""
        return self._is_empty

    def go_to_next_file(self):
        """Increments the current file/file index to the next file. Raises EmptyError if there are no more files."""
        if self.current_file_index == len(self.files) - 1:
            self._is_empty = True
            raise EmptyError()
        else:
            self.current_file_index += 1
            self.current_file.close()
            self.current_file = open(self.files[self.current_file_index], "rb")

    def pop(self, size):
        """Gets size bytes from the queue"""
        b = bytearray()

        if self.current_file is None:
            self.current_file = open(self.files[self.current_file_index], "rb")

        while True:
            b.extend(self.current_file.read(size - len(b)))
            if len(b) < size:
                try:
                    self.go_to_next_file()
                except EmptyError:
                    return b
            else:
                return b


def events_from_files_generator(files, buffer_size=int(1e9)):
    """Yields events from a list of binary files"""
    reader = ListModeDataReader()
    bytes_stream = FilesIO(files)
    while not bytes_stream.is_empty():
        reader.put(bytes_stream.pop(buffer_size))
        yield from reader.iterevents()
    if not reader.is_empty():
        print("[orange3]WARNING[/] Left over bytes in binary data stream.")


def to_list(obj):
    if isinstance(obj, str):
        obj = [obj]
    else:
        try:
            obj = list(obj)
        except TypeError:
            obj = [obj]
    return obj


def read_list_mode_data(files, buffer_size=int(1e9)):
    """Loads a list of binary files into a pandas DataFrame"""
    files = to_list(files)
    gen = events_from_files_generator(files, buffer_size)
    events = tuple(gen)
    return pd.DataFrame.from_records(events, columns=Event._fields)


def read_list_mode_data_as_events(files, buffer_size=int(1e9), max_size=None):
    """Returns a list of at most max_size events from files"""
    files = to_list(files)
    gen = events_from_files_generator(files, buffer_size)
    return list(islice(gen, max_size))



"""pixie16-binary-browser for *.bin files.

Usage:
  pixie16-binary-browser [<filename>]

"""

from docopt import docopt

commands = docopt(__doc__)


def main():
    from pathlib import Path
    import sys

    from pixie16.ui import BinaryApp, QApplication

    filename = commands["<filename>"]
    filename = Path(filename) if filename else None
    if filename and not filename.exists():
        print(f"[Error] Cannot find file {filename}")
        sys.exit(1)

    app = QApplication([])
    ex = BinaryApp(app)
    if filename:
        ex.open_file(filename)
    sys.exit(app.exec_())



from pathlib import Path

import pytest

import pixie16


@pytest.fixture
def setting():
    return pixie16.read.Settings.from_file(Path(__file__).parent / "settings_file.json")


@pytest.fixture
def stats():
    return pixie16.read.Stats.from_file(Path(__file__).parent / "stats_file.json")


class TestSettings:
    def test_standard_access(self, setting):
        assert setting["channel"]["input"]["BLcut"][0] == 223
        assert setting["channel"]["input"]["BLcut"][1] == 242

    def test_flat_access(self, setting):
        assert setting.get_by_name("BLcut")[0] == 223
        assert setting.get_by_name("BLcut")[1] == 242

    def test_setting_value(self, setting):
        setting["channel"]["input"]["BLcut"][0] = 2
        assert setting["channel"]["input"]["BLcut"][0] == 2

    def test_setting_by_flat_access(self, setting):
        setting.set_by_name("BLcut", list(range(16)))
        assert setting["channel"]["input"]["BLcut"][0] == 0
        assert setting["channel"]["input"]["BLcut"][4] == 4


class TestStats:
    def test_standard_access(self, stats):
        assert stats["real_time"][0] == 1.0
        assert stats["real_time"][1] == 2.0

    def test_flat_access(self, stats):
        assert stats.get_by_name("real_time")[0] == 1.0
        assert stats.get_by_name("real_time")[1] == 2.0
